class MasterIO(object):

    def __init__(self, filename):
        """Create an IO object attached to a file.

        :param str filename: The file to read from or write to
        """
        raise NotImplementedError()

    def __enter__(self):
        raise NotImplementedError()

    def write(self, obj):
        """Write one or more objects to the file.

        :param obj: The object(s) to write
        :type obj: [:class:`splipy.SplineObject`] or :class:`splipy.SplineObject`
        """
        raise NotImplementedError()

    def read(self):
        """Reads all the objects from the file.

        :return: Objects
        :rtype: [:class:`splipy.SplineObject`]
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-

import numpy as np
import copy
from operator import attrgetter, methodcaller
from itertools import chain, product
from bisect import bisect_left

from .basis import BSplineBasis
from .utils import (
    reshape, rotation_matrix, is_singleton, ensure_listlike,
    check_direction, ensure_flatlist, check_section, sections,
    raise_order_1D
)

__all__ = ['SplineObject']


def transpose_fix(pardim, direction):
    ret = list(range(1, pardim+1))
    ret.insert(direction, 0)
    return tuple(ret)


def evaluate(bases, cps, tensor=True):
    if tensor:
        idx = len(bases) - 1
        for N in bases[::-1]:
            cps = np.tensordot(N, cps, axes=(1, idx))
    else:
        cps = np.einsum('ij,j...->i...', bases[0], cps)
        for N in bases[1:]:
            cps = np.einsum('ij,ij...->i...', N, cps)
    return cps


class SplineObject(object):
    """  Master class for spline objects with arbitrary dimensions.

    This class should be subclassed instead of used directly.

    All SplineObjects support basic arithmetic operators, which are interpreted
    as translation and scaling. In-place operators (e.g. ``+=``) mutate the
    object, while infix operators (e.g. ``+``) create new objects.
    """

    def __init__(self, bases=None, controlpoints=None, rational=False, raw=False):
        """  Construct a spline object with the given bases and control points.

        The default is to create a linear one-element mapping from and to the
        unit (hyper)cube.

        :param [BSplineBasis] bases: The basis of each parameter direction
        :param array-like controlpoints: An *n1* × *n2* × ... × *d* matrix of
            control points
        :param bool rational: Whether the object is rational (in which case the
            control points are interpreted as pre-multiplied with the weight,
            which is the last coordinate)
        :param bool raw: If True, skip any control point reordering.
            (For internal use.)
        """
        bases = [(b.clone() if b else BSplineBasis()) for b in bases]
        self.bases = bases
        if controlpoints is None:
            # `product' produces tuples in row-major format (the last input varies quickest)
            # We want them in column-major format, so we reverse the basis orders, and then
            # also reverse the output tuples
            controlpoints = [c[::-1] for c in product(*(b.greville() for b in bases[::-1]))]

            # Minimum two dimensions
            if len(controlpoints[0]) == 1:
                controlpoints = [tuple(list(c) + [0.0]) for c in controlpoints]

            # Add weight = 1 for identiy-mapping rational splines
            if rational:
                controlpoints = [tuple(list(c) + [1.0]) for c in controlpoints]

        self.controlpoints = np.array(controlpoints)
        self.dimension = self.controlpoints.shape[-1] - rational
        self.rational = rational

        if not raw:
            shape = tuple(b.num_functions() for b in bases)
            ncomps = self.dimension + rational
            self.controlpoints = reshape(self.controlpoints, shape, order='F', ncomps=ncomps)

    def _validate_domain(self, *params):
        """  Check whether the given evaluation parameters are valid.

        :raises ValueError: If the parameters are outside the domain
        """
        for b, p in zip(self.bases, params):
            b.snap(p)
            if b.periodic < 0:
                if min(p) < b.start() or b.end() < max(p):
                    raise ValueError('Evaluation outside parametric domain')

    def evaluate(self, *params, **kwargs):
        """  Evaluate the object at given parametric values.

        If *tensor* is true, evaluation will take place on a tensor product
        grid, i.e. it will return an *n1* × *n2* × ... × *dim* array, where
        *ni* is the number of evaluation points in direction *i*, and *dim* is
        the physical dimension of the object.

        If *tensor* is false, there must be an equal number *n* of evaluation
        points in all directions, and the return value will be an *n* × *dim*
        array.

        If there is only one evaluation point, a vector of length *dim* is
        returned instead.

        :param u,v,...: Parametric coordinates in which to evaluate
        :type u,v,...: float or [float]
        :param tensor: Whether to evaluate on a tensor product grid
        :type tensor: bool
        :return: Geometry coordinates
        :rtype: numpy.array
        """
        squeeze = all(is_singleton(p) for p in params)
        params = [ensure_listlike(p) for p in params]

        tensor = kwargs.get('tensor', True)
        if not tensor and len({len(p) for p in params}) != 1:
            raise ValueError('Parameters must have same length')

        self._validate_domain(*params)

        # Evaluate the corresponding bases at the corresponding points
        # and build the result array
        Ns = [b.evaluate(p) for b, p in zip(self.bases, params)]
        result = evaluate(Ns, self.controlpoints, tensor)

        # For rational objects, we divide out the weights, which are stored in the
        # last coordinate
        if self.rational:
            for i in range(self.dimension):
                result[..., i] /= result[..., -1]
            result = np.delete(result, self.dimension, -1)

        # Squeeze the singleton dimensions if we only have one point
        if squeeze:
            result = result.reshape(self.dimension)

        return result

    def derivative(self, *params, **kwargs):
        """  Evaluate the derivative of the object at the given parametric values.

        If *tensor* is true, evaluation will take place on a tensor product
        grid, i.e. it will return an *n1* × *n2* × ... × *dim* array, where
        *ni* is the number of evaluation points in direction *i*, and *dim* is
        the physical dimension of the object.

        If *tensor* is false, there must be an equal number *n* of evaluation
        points in all directions, and the return value will be an *n* × *dim*
        array.

        If there is only one evaluation point, a vector of length *dim* is
        returned instead.

        Examples:

        .. code:: python

           # Tangent of curve at single point
           curve.derivative(1.0)

           # Double derivative of curve at single point:
           curve.derivative(1.0, d=2)

           # Third derivative of curve at several points:
           curve.derivative([0.0, 1.0, 2.0], d=3)

           # Tangents of surface:
           surface.derivative(0.5, 0.7, d=(1,0))
           surface.derivative(0.5, 0.7, d=(0,1))

           # Cross-derivative of surface:
           surface.derivative(0.5, 0.7, d=(1,1))

        :param u,v,...: Parametric coordinates in which to evaluate
        :type u,v,...: float or [float]
        :param (int) d: Order of derivative to compute
        :param (bool) above: Evaluation in the limit from above
        :param tensor: Whether to evaluate on a tensor product grid
        :type tensor: bool
        :return: Derivatives
        :rtype: numpy.array
        """
        squeeze = all(is_singleton(p) for p in params)
        params = [ensure_listlike(p) for p in params]

        derivs = kwargs.get('d', [1] * self.pardim)
        derivs = ensure_listlike(derivs, self.pardim)

        above = kwargs.get('above', [True] * self.pardim)
        above = ensure_listlike(above, self.pardim)

        tensor = kwargs.get('tensor', True)

        if not tensor and len({len(p) for p in params}) != 1:
            raise ValueError('Parameters must have same length')

        self._validate_domain(*params)

        # Evaluate the derivatives of the corresponding bases at the corresponding points
        # and build the result array
        dNs = [b.evaluate(p, d, from_right) for b, p, d, from_right in zip(self.bases, params, derivs, above)]
        result = evaluate(dNs, self.controlpoints, tensor)

        # For rational curves, we need to use the quotient rule
        # (n/W)' = (n' W - n W') / W^2 = n'/W - nW'/W^2
        # * n'(i) = result[..., i]
        # * W'(i) = result[..., -1]
        # We evaluate in the regular way to compute n and W.
        if self.rational:
            if sum(derivs) > 1:
                raise RuntimeError('Rational derivative not implemented for order %i' % sum(derivs))
            Ns = [b.evaluate(p) for b, p in zip(self.bases, params)]
            non_derivative = evaluate(Ns, self.controlpoints, tensor)
            W = non_derivative[..., -1]  # W
            Wd = result[..., -1]         # W'
            for i in range(self.dimension):
                result[..., i] = result[..., i] / W - non_derivative[..., i] * Wd / W / W
            result = np.delete(result, self.dimension, -1)

        # Squeeze the singleton dimensions if we only have one point
        if squeeze:
            result = result.reshape(self.dimension)

        return result

    def get_derivative_spline(self, direction=None):
        """  Compute the controlpoints associated with the derivative spline object

        If `direction` is given, only the derivatives in that direction are
        returned.

        If `direction` is not given, this function returns a tuple of all
        partial derivatives

        .. code:: python

           # Create a 4x4 element cubic spline surface
           surf = Surface()
           surf.raise_order(2,2)
           surf.refine(3,3)
           surf[1:4,1:4,:] += 0.1 # make the surface non-trivial by moving controlpoints

           # Create the derivative surface
           du = surf.get_derivative_spline(direction='u')

           # evaluation is identical
           print(du.evaluate(0.3, 0.4))
           print(surf.derivative(0.3, 0.4, d=(1,0)))

           print(surf.order()) # prints (3,3)
           print(du.order())   # prints (2,3)

        :param int direction: The tangential direction
        :return: Derivative spline
        :rtype: SplineObject
        """

        if self.rational:
            raise RuntimeError('Not working for rational splines')

        # if no direction is specified, return a tuple with all derivatives
        if direction is None:
            return tuple([self.get_derivative_spline(dim) for dim in range(self.pardim)])
        else:
            d = check_direction(direction, self.pardim)
            k = self.knots(d, with_multiplicities=True)
            p = self.order(d)-1
            n = self.shape[d]
            if self.bases[d].periodic < 0:
                C = np.zeros((n-1, n))
                for i in range(n-1):
                    C[i,i]   = -float(p) / (k[i+p+1] - k[i+1])
                    C[i,i+1] =  float(p) / (k[i+p+1] - k[i+1])
            else:
                C = np.zeros((n, n))
                for i in range(n):
                    ip1 = np.mod(i+1,n)
                    C[i,i]   = -float(p) / (k[i+p+1] - k[i+1])
                    C[i,ip1] =  float(p) / (k[i+p+1] - k[i+1])

            derivative_cps = np.tensordot(C, self.controlpoints, axes=(1, d))
            derivative_cps = derivative_cps.transpose(transpose_fix(self.pardim, d))
            bases    = [b for b in self.bases]
            bases[d] = BSplineBasis(p, k[1:-1], bases[d].periodic-1)

            # search for the right subclass constructor, i.e. Volume, Surface or Curve
            constructor = [c for c in SplineObject.__subclasses__() if c._intended_pardim == len(self.bases)]
            constructor = constructor[0]

            # return derivative object
            args = bases + [derivative_cps] + [self.rational]
            return constructor(*args, raw=True)


    def tangent(self, *params, **kwargs):
        """  Evaluate the tangents of the object at the given parametric values.

        If `direction` is given, only the derivatives in that direction are
        evaluated. This is equivalent to calling
        :func:`splipy.SplineObject.derivative` with
        `d=(0,...,0,1,0,...,0)`, the unit vector corresponding to the given
        direction.

        If `direction` is not given, this function returns a tuple of all
        tangential derivatives at the given points.

        :param u,v,...: Parametric coordinates in which to evaluate
        :type u,v,...: float or [float]
        :param int direction: The tangential direction
        :param (bool) above: Evaluation in the limit from above
        :param tensor: Whether to evaluate on a tensor product grid
        :type tensor: bool
        :return: Tangents
        :rtype: tuple<numpy.array>
        """
        direction = kwargs.get('direction', None)
        derivative = [0] * self.pardim

        above = kwargs.get('above', [True] * self.pardim)
        above = ensure_listlike(above, self.pardim)

        tensor = kwargs.get('tensor', True)

        if self.pardim == 1: # curves
            direction = 0

        if direction is None:
            result = ()
            for i in range(self.pardim):
                derivative[i] = 1
                # compute velocity in this direction
                v = self.derivative(*params, d=derivative, above=above, tensor=tensor)
                # normalize
                if len(v.shape)==1:
                    speed = np.linalg.norm(v)
                else:
                    speed = np.linalg.norm( v, axis=-1)
                    speed = np.reshape(speed, speed.shape +(1,))
                # store in result tuple
                result += (v/speed,)
                derivative[i] = 0
            return result

        i = check_direction(direction, self.pardim)
        derivative[i] = 1
        # compute velocity in this direction
        v = self.derivative(*params, d=derivative, above=above, tensor=tensor)
        # normalize
        if len(v.shape)==1:
            speed = np.linalg.norm(v)
        else:
            speed = np.linalg.norm( v, axis=-1)
            speed = np.reshape(speed, speed.shape +(1,))

        return v / speed

    def section(self, *args, **kwargs):
        """  Returns a section from the object. A section can be any sub-object of
        parametric dimension not exceeding that of the object. E.g. for a
        volume, sections include vertices, edges, faces, etc.

        The arguments are control point indices for each direction. `None`
        means that direction should be variable in the returned object.

        .. code:: python

           # Get the face with u=max
           vol.section(-1, None, None)

           # Keyword arguments are supported for u, v and w
           # This is the same as the above
           vol.section(u=-1)

           # Get the edge with u=min, v=max
           vol.section(0, -1, None)

           # This is equivalent to vol.clone()
           vol.section()

        If a specific subclass of `SplineObject` is found that handles the
        requested number of variable directions (parametric dimension), then
        the return value is of that type. If not, it will be a generic `SplineObject`.

        If the section has no variable directions (it is a point), then the
        return value will be an array, unless the keyword argument
        `unwrap_points` is true, in which case it will return a
        zero-dimensional `SplineObject`.

        :param u,v,...: Control point indices
        :type u,v,...: int or None
        :return: Section
        :rtype: SplineObject or np.array
        """
        section = check_section(*args, pardim=self.pardim, **kwargs)
        unwrap_points = kwargs.get('unwrap_points', True)

        slices = tuple(slice(None) if p is None else p for p in section)
        bases = [b for b, p in zip(self.bases, section) if p is None]
        if bases or not unwrap_points:
            classes = [c for c in SplineObject.__subclasses__() if c._intended_pardim == len(bases)]
            if classes:
                args = bases + [self.controlpoints[slices], self.rational]
                return classes[0](*args, raw=True)
            return SplineObject(bases, self.controlpoints[slices], self.rational, raw=True)
        return self.controlpoints[slices]

    def set_order(self, *order):
        """  Set the polynomial order of the object. If only one argument is
        given, the order is set uniformly over all directions.

        :param int u,v,...: The new order in a given direction.
        :raises ValueError: If the order is reduced in any direction.
        :return: self
        """
        if len(order) == 1:
            order = [order[0]] * self.pardim
        if not all(new >= old for new, old in zip(order, self.order())):
            raise ValueError("Cannot lower order using set_order")

        diff = [new - old for new, old in zip(order, self.order())]
        return self.raise_order(*diff)

    def raise_order(self, *raises, direction=None):
        """  Raise the polynomial order of the object. If only one
        argument is given, the order is raised equally over all
        directions, unless the `direction` argument is also given. The
        explicit version is only implemented on open knot vectors. The
        function raise_order_implicit is used otherwise.

        :param int u,v,...: Number of times to raise the order in a given
            direction.
        :param int direction: The direction in which to raise the order.
        :return: self
        """
        if len(raises) == 1 and direction is None:
            raises = [raises[0]] * self.pardim
        elif len(raises) == 1:
            newraises = [0] * self.pardim
            newraises[check_direction(direction, self.pardim)] = raises[0]
            raises = newraises
        if not all(r >= 0 for r in raises):
            raise ValueError("Cannot lower order using raise_order")
        if all(r == 0 for r in raises):
            return self

        if any(b.continuity(b.knots[0]) < b.order or b.periodic > -1 for b in self.bases):
            self.raise_order_implicit(*raises)
            return self

        new_bases = [b.raise_order(r) for b, r in zip(self.bases, raises)]

        d_p = self.pardim

        controlpoints = self.controlpoints
        for i in range(0,d_p):
            dimensions = np.array(controlpoints.shape)
            indices = np.array(range(0,d_p+1))
            indices[i],indices[d_p] = d_p,i
            controlpoints = np.transpose(controlpoints,indices)
            controlpoints = np.reshape(controlpoints,(np.prod(dimensions[indices[:-1]]),dimensions[i]))
            controlpoints = raise_order_1D(controlpoints.shape[1]-1,self.order(i),
                                           self.bases[i].knots,controlpoints,raises[i],self.bases[i].periodic)
            controlpoints = np.reshape(controlpoints,np.append(dimensions[indices[:-1]],controlpoints.shape[1]))
            controlpoints = np.transpose(controlpoints,indices)

        self.controlpoints = controlpoints
        self.bases = new_bases
        return self

    def raise_order_implicit(self, *raises):
        """  Raise the polynomial order of the object. If only one argument is
        given, the order is raised equally over all directions.

        :param int u,v,...: Number of times to raise the order in a given
            direction.
        :return: self
        """

        new_bases = [b.raise_order(r) for b, r in zip(self.bases, raises)]

        # Set up an interpolation problem
        # This works in projective space, so no special handling for rational objects
        interpolation_pts = [b.greville() for b in new_bases]
        N_old = [b(pts) for b, pts in zip(self.bases, interpolation_pts)]
        N_new = [b(pts) for b, pts in zip(new_bases, interpolation_pts)]

        # Calculate the projective interpolation points
        result = self.controlpoints
        for n in N_old[::-1]:
            result = np.tensordot(n, result, axes=(1, self.pardim-1))

        # Solve the interpolation problem
        for n in N_new[::-1]:
            result = np.tensordot(np.linalg.inv(n), result, axes=(1, self.pardim-1))

        self.controlpoints = result
        self.bases = new_bases

        return self

    def lower_order(self, *lowers):
        """  Lower the polynomial order of the object. If only one argument is
        given, the order is lowered equally over all directions.

        :param int u,v,...: Number of times to lower the order in a given
            direction.
        :return SplineObject: Approximation of the current object on a lower
            order basis
        """
        if len(lowers) == 1:
            lowers = [lowers[0]] * self.pardim
        if all(l == 0 for l in lowers):
            return self.clone()

        new_bases = [b.lower_order(l) for b, l in zip(self.bases, lowers)]

        # Set up an interpolation problem
        # This works in projective space, so no special handling for rational objects
        interpolation_pts = [b.greville() for b in new_bases]
        N_old = [b(pts) for b, pts in zip(self.bases, interpolation_pts)]
        N_new = [b(pts) for b, pts in zip(new_bases, interpolation_pts)]

        # Calculate the projective interpolation points
        new_controlpts = self.controlpoints
        for n in N_old[::-1]:
            new_controlpts = np.tensordot(n, new_controlpts, axes=(1, self.pardim-1))

        # Solve the interpolation problem
        for n in N_new[::-1]:
            new_controlpts = np.tensordot(np.linalg.inv(n), new_controlpts, axes=(1, self.pardim-1))

        # search for the right subclass constructor, i.e. Volume, Surface or Curve
        constructor = [c for c in SplineObject.__subclasses__() if c._intended_pardim == len(self.bases)]
        constructor = constructor[0]

        # return approximated object
        args = new_bases + [new_controlpts] + [self.rational]
        return constructor(*args, raw=True)

    def start(self, direction=None):
        """  Return the start of the parametric domain.

        If `direction` is given, returns the start of that direction, as a
        float. If it is not given, returns the start of all directions, as a
        tuple.

        :param int direction: Direction in which to get the start.
        :raises ValueError: For invalid direction
        """
        if direction is None:
            return tuple(b.start() for b in self.bases)
        direction = check_direction(direction, self.pardim)
        return self.bases[direction].start()

    def end(self, direction=None):
        """  Return the end of the parametric domain.

        If `direction` is given, returns the end of that direction, as a float.
        If it is not given, returns the end of all directions, as a tuple.

        :param int direction: Direction in which to get the end.
        :raises ValueError: For invalid direction
        """
        if direction is None:
            return tuple(b.end() for b in self.bases)
        direction = check_direction(direction, self.pardim)
        return self.bases[direction].end()

    def order(self, direction=None):
        """  Return polynomial order (degree + 1).

        If `direction` is given, returns the order of that direction, as an
        int. If it is not given, returns the order of all directions, as a
        tuple.

        :param int direction: Direction in which to get the order.
        :raises ValueError: For invalid direction
        """
        if direction is None:
            return tuple(b.order for b in self.bases)
        direction = check_direction(direction, self.pardim)
        return self.bases[direction].order

    def knots(self, direction=None, with_multiplicities=False):
        """  Return knots vector

        If `direction` is given, returns the knots in that direction, as a
        list. If it is not given, returns the knots of all directions, as a
        tuple.

        :param int direction: Direction in which to get the knots.
        :param bool with_multiplicities: If true, return knots with
            multiplicities (i.e. repeated).
        :raises ValueError: For invalid direction
        """
        getter = attrgetter('knots') if with_multiplicities else methodcaller('knot_spans')
        if direction is None:
            return tuple(getter(b) for b in self.bases)
        direction = check_direction(direction, self.pardim)
        return getter(self.bases[direction])

    def reverse(self, direction=0):
        """  Swap the direction of a parameter by making it go in the reverse
        direction. The parametric domain remains unchanged.

        :param int direction: The direction to flip.
        :return: self
        """
        direction = check_direction(direction, self.pardim)
        self.bases[direction].reverse()

        # This creates the following slice programmatically
        # array[:, :, :, ..., ::-1,]
        # index=direction -----^
        # :    => slice(None, None, None)
        # ::-1 => slice(None, None, -1)
        direction = check_direction(direction, self.pardim)
        slices = [slice(None, None, None) for _ in range(direction)] + [slice(None, None, -1)]
        self.controlpoints = self.controlpoints[tuple(slices)]

        return self

    def swap(self, dir1=0, dir2=1):
        """  Swaps two parameter directions.

        This function silently passes for curves.

        :param direction dir1: The first direction (default u)
        :param direction dir2: The second direction (default v)
        :return: self
        """
        if self.pardim == 1:
            return

        dir1 = check_direction(dir1, self.pardim)
        dir2 = check_direction(dir2, self.pardim)

        # Reorder control points
        new_directions = list(range(self.pardim + 1))
        new_directions[dir1] = dir2
        new_directions[dir2] = dir1
        self.controlpoints = self.controlpoints.transpose(new_directions)

        # Swap knot vectors
        self.bases[dir1], self.bases[dir2] = self.bases[dir2], self.bases[dir1]

        return self

    def insert_knot(self, knot, direction=0):
        """  Insert a new knot into the spline.

        :param int direction: The direction to insert in
        :param knot: The new knot(s) to insert
        :type knot: float or [float]
        :raises ValueError: For invalid direction
        :return: self
        """
        shape  = self.controlpoints.shape

        # for single-value input, wrap it into a list
        knot = ensure_listlike(knot)

        direction = check_direction(direction, self.pardim)

        C = np.identity(shape[direction])
        for k in knot:
            C = self.bases[direction].insert_knot(k) @ C
        self.controlpoints = np.tensordot(C, self.controlpoints, axes=(1, direction))
        self.controlpoints = self.controlpoints.transpose(transpose_fix(self.pardim, direction))

        return self

    def refine(self, *ns, **kwargs):
        """  Enrich the spline space by inserting knots into each existing knot
        span.

        This method supports three different usage patterns:

        .. code:: python

           # Refine each direction by a factor n
           obj.refine(n)

           # Refine a single direction by a factor n
           obj.refine(n, direction='v')

           # Refine all directions by given factors
           obj.refine(nu, nv, ...)

        :param int nu,nv,...: Number of new knots to insert into each span
        :param int direction: Direction to refine in
        :return: self
        """
        direction = kwargs.get('direction', None)

        if len(ns) == 1 and direction is not None:
            directions = [check_direction(direction, self.pardim)]
        else:
            directions = range(self.pardim)

        if len(ns) == 1:
            ns = [ns[0]] * self.pardim

        for n, d in zip(ns, directions):
            knots = self.knots(direction=d)  # excluding multiple knots
            new_knots = []
            for (k0, k1) in zip(knots[:-1], knots[1:]):
                new_knots.extend(np.linspace(k0, k1, n+2)[1:-1])
            self.insert_knot(new_knots, d)

        return self

    def reparam(self, *args, **kwargs):
        """  Redefine the parametric domain. This function accepts two calling
        conventions:

        `reparametrize(u, v, ...)` reparametrizes each direction to the domains
        given by the tuples *u*, *v*, etc. It is equivalent to calling
        `reparametrize(u[0], u[1])` on each basis. The default domain for
        directions not given is (0,1). In particular, if no arguments are
        given, the new parametric domain will be the unit (hyper)cube.

        `reparametrize(u, direction=d)` reparametrizes just the direction given
        by *d* and leaves the others untouched.

        :param tuple u, v, ...: New parametric domains, default to (0,1)
        :param int direction: The direction to reparametrize
        :return: self
        """
        if 'direction' not in kwargs:
            # Pad the args with (0, 1) for the extra directions
            args = list(args) + [(0, 1)] * (len(self.bases) - len(args))
            for b, (start, end) in zip(self.bases, args):
                b.reparam(start, end)
        else:
            direction = kwargs['direction']
            direction = check_direction(direction, self.pardim)
            if len(args) == 0:
                self.bases[direction].reparam(0,1)
            else:
                start, end = args[0]
                self.bases[direction].reparam(start, end)

        return self

    def translate(self, x):
        """  Translate (i.e. move) the object by a given distance.

        :param array-like x: The vector to translate by.
        :return: self
        """
        # 3D rational example: create a 4x4 translation matrix
        #
        #  |xw|      |  1   0   0  x1 |   |xw|
        #  |yw|   =  |  0   1   0  x2 | * |yw|
        #  |zw|      |  0   0   1  x3 |   |zw|
        #  | w|_new  |  0   0   0   1 |   | w|_old
        #
        #  PS: we even need a rational representation for non-rational splines
        #      in order to formulate translation as a matrix-matrix product
        dim = self.dimension
        rat = self.rational
        n = len(self)  # number of control points
        if len(x) > dim:  # typical case: requesting movement in z-direction for 2D geometries
            self.set_dimension(len(x))
            dim = self.dimension

        # set up the translation matrix
        translation_matrix = np.identity(dim + 1)
        for i in range(dim):
            translation_matrix[i, -1] = x[i]

        # wrap out the controlpoints to a matrix (down from n-D tensor)
        if not self.rational:
            cp = np.ones((n, dim + 1))  # pad with weights=1
            cp[:, :-1] = np.reshape(self.controlpoints, (n, dim))
        else:
            cp = np.reshape(self.controlpoints, (n, dim + rat))

        # do the actual scaling by matrix-matrix multiplication
        cp = cp @ translation_matrix.T  # right-mult, so we need transpose

        # store results
        if self.rational:
            self.controlpoints = np.reshape(np.array(cp), self.controlpoints.shape)
        else:
            self.controlpoints = np.reshape(np.array(cp[:, :-1]), self.controlpoints.shape)

        return self

    def scale(self, *args):
        """  Scale, or magnify the object by a given amount.

        In case of one input argument, the scaling is uniform.

        :param args: Scaling factors, possibly different in each direction.
        :type args: array-like or float
        :return: self
        """
        # 3D rational example: create a 4x4 scaling matrix
        #
        #  |xw|      |  sx    0    0   0 |   |xw|
        #  |yw|   =  |   0   sy    0   0 | * |yw|
        #  |zw|      |   0    0   sz   0 |   |zw|
        #  | w|_new  |   0    0    0   1 |   | w|_old
        #
        dim = self.dimension
        rat = self.rational
        n = len(self)  # number of control points
        s = ensure_flatlist(args)
        s = ensure_listlike(s, dups=3)

        # set up the scaling matrix
        scale_matrix = np.identity(dim + rat)
        for i in range(dim):
            scale_matrix[i, i] = s[i]

        # wrap out the controlpoints to a matrix (down from n-D tensor)
        cp = np.reshape(self.controlpoints, (n, dim + rat))

        # do the actual scaling by matrix-matrix multiplication
        cp = cp @ scale_matrix

        # store results
        self.controlpoints = np.reshape(np.array(cp), self.controlpoints.shape)

        return self

    def rotate(self, theta, normal=(0, 0, 1)):
        """  Rotate the object around an axis.

        :param float theta: Angle to rotate about, measured in radians
        :param array-like normal: The normal axis (if 3D) to rotate about
        :raises RuntimeError: If the physical dimension is not 2 or 3
        :return: self
        """
        # 2D rational example: create a 3x3 rotation matrix
        #
        #  |xw|     | cos(t)  -sin(t)   0 |   |xw|
        #  |yw|   = | sin(t)   cos(t)   0 | * |yw|
        #  | w|_new |   0        0      1 |   | w|_old
        #

        dim = self.dimension
        rat = self.rational
        n = len(self)  # number of control points
        if not (normal[0] == 0 and normal[1] == 0):  # rotating a 2D geometry out of the xy-plane
            self.set_dimension(3)
            dim = self.dimension

        # set up the rotation matrix
        if dim == 2:
            R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]
                 ]).T  # we do right-multiplication, so we need a transpose
        elif dim == 3:
            normal = np.array(normal)
            R = rotation_matrix(theta, normal)
        else:
            raise RuntimeError('rotation undefined for geometries other than 2D and 3D')

        rot_matrix = np.identity(dim + rat)
        rot_matrix[0:dim, 0:dim] = R

        # wrap out the controlpoints to a matrix (down from n-D tensor)
        cp = np.reshape(self.controlpoints, (n, dim + rat))

        # do the actual rotation by matrix-matrix multiplication
        cp = cp @ rot_matrix

        # store results
        self.controlpoints = np.reshape(np.array(cp), self.controlpoints.shape)

        return self

    def mirror(self, normal):
        """  Mirror the object around a plane through the origin.

        :param array-like normal: The plane normal to mirror about.
        :raises RuntimeError: If the physical dimension is not 2 or 3
        :return: self
        """
        # 3D rational example: create a 4x4 reflection matrix
        #
        #  normal = [a,b,c]
        #
        #  |xw|     |  1-2a^2   -2ab    -2ac   0 |   |xw|
        #  |yw|   = |   -2ab   1-2b^2   -2bc   0 | * |yw|
        #  |zw|   = |   -2ac    -2bc   1-2c^2  0 | * |zw|
        #  | w|_new |    0       0         0   1 |   | w|_old
        #
        #  PS: A reflection about a line is not a linear transformation; it is
        #      an affine transformation.

        dim = self.dimension
        rat = self.rational
        n = len(self)  # number of control points

        if dim != 3:
            raise RuntimeError('reflection undefined for geometries other than 3D')

        # fixup the input normal to right form
        normal = np.array(normal)
        normal = normal / np.sqrt(np.dot(normal, normal))  # normalize it

        # set up the reflection matrix
        reflection_matrix = np.identity(dim + rat)
        reflection_matrix[0:dim, 0:dim] -= 2 * np.outer(normal, normal)

        # wrap out the controlpoints to a matrix (down from n-D tensor)
        cp = np.reshape(self.controlpoints, (n, dim + rat))

        # do the actual rotation by matrix-matrix multiplication
        cp = cp @ reflection_matrix

        # store results
        self.controlpoints = np.reshape(np.array(cp), self.controlpoints.shape)

        return self

    def project(self, plane):
        """  Projects the geometry onto a plane or axis.

        - `project('xy')` will project the object onto the *xy* plane, setting
          all *z* components to zero.
        - `project('y')` will project the object onto the *y* axis, setting all
          *x* and *z* components to zero.

        :param string plane: Any combination of 'x', 'y' and 'z'
        :return: self
        """
        keep = [c in plane.lower() for c in 'xyz']

        dim = self.dimension
        for i in range(dim):
            if not keep[i]:
                self.controlpoints[..., i] = 0

        return self

    def bounding_box(self):
        """  Gets the bounding box of a spline object, computed from the
        control-point values. Could be inaccurate for rational splines.

        Returns the minima and maxima for each direction:
        [(xmin, xmax), (ymin, ymax), ...]

        :return: Bounding box
        :rtype: [(float)]
        """
        dim = self.dimension

        result = []
        for i in range(dim):
            result.append((np.min(self.controlpoints[..., i]),
                           np.max(self.controlpoints[..., i])))
        return result

    def center(self):
        """  Gets the center of the domain

        For curves this will return :math:`(\\tilde{x}, \\tilde{y},...)`, where

        .. math:: \\tilde{x} = \\frac{1}{L} \\int_{t_0}^{t_1} x(t) \\; dt

        and :math:`L=t_1-t_0` is the length of the parametric domain :math:`[t_0,t_1]`.

        For surfaces this will return :math:`(\\tilde{x}, \\tilde{y},...)`, where

        .. math:: \\tilde{x} = \\frac{1}{A} \\int_{v_0}^{v_1} \\int_{u_0}^{u_1} x(u,v) \\; du \\; dv

        and :math:`A=(u_1-u_0)(v_1-v_0)` is the area of the parametric domain :math:`[u_0,u_1]\\times[v_0,v_1]`.

        .. warning:: For rational splines, this will integrate in projective
            coordinates, then project the centerpoint. This is as opposed to
            integrating the rational functions :math:`\\frac{N_i(t)w_i}{\\sum_j
            N_j(t)w_j}`.
        """

        # compute integration of basis functions
        Ns = [b.integrate(b.start(), b.end()) for b in self.bases]

        # compute parametric size
        par_size = np.prod([t1-t0 for (t0,t1) in zip(self.start(), self.end())])

        # multiply basis functions with control points
        idx = self.pardim - 1
        result = self.controlpoints
        for N in Ns[::-1]:
            result = np.tensordot(N, result, axes=(0, idx))
            idx -= 1

        result /= par_size

        # project to physical space
        if self.rational:
            result[:-1] /= result[-1]
            result = np.delete(result, self.dimension)

        return result

    def corners(self, order='C'):
        """  Return the corner control points.

        The `order` parameter determines which order to use, either ``'F'`` or
        ``'C'``, for row-major or column-major ordering. E.g. for a volume, in
        parametric coordinates,

        - ``'C'`` gives (0,0,0), (1,0,0), (0,1,0), (1,1,0), (0,0,1), etc.
        - ``'F'`` gives (0,0,0), (0,0,1), (0,1,0), (0,1,1), (1,0,0), etc.

        :param str order: The ordering to use
        :return: Corners
        :rtype: np.array

        .. warning:: For rational splines, this will return the corners in
            projective coordinates, including weights.
        """
        result = np.zeros((2**self.pardim, self.dimension + int(self.rational)))
        for i, args in enumerate(sections(self.pardim, 0)):
            result[i,:] = self.section(*(args[::-1] if order == 'F' else args))
        return result

    def lower_periodic(self, periodic, direction=0):
        """  Sets the periodicity of the spline object in the given direction,
        keeping the geometry unchanged.

        :param int periodic: new periodicity, i.e. the basis is C^k over the start/end
        :param int direction: the parametric direction of the basis to modify
        :return: self
        """
        direction = check_direction(direction, self.pardim)

        b  = self.bases[direction]
        while periodic < b.periodic:
            self.insert_knot(self.start(direction), direction)
            self.controlpoints = np.roll(self.controlpoints, -1, direction)
            b.roll(1)
            b.periodic -= 1
            b.knots = b.knots[:-1]
        if periodic > b.periodic:
            raise ValueError('Cannot raise periodicity')

        return self

    def set_dimension(self, new_dim):
        """  Sets the physical dimension of the object. If increased, the new
        components are set to zero.

        :param int new_dim: New dimension.
        :return: self
        """
        dim = self.dimension
        shape = self.controlpoints.shape
        while new_dim > dim:
            self.controlpoints = np.insert(self.controlpoints, dim, np.zeros(shape[:-1]), self.pardim)
            dim += 1
        while new_dim < dim:
            self.controlpoints = np.delete(self.controlpoints, -2 if self.rational else -1, -1)
            dim -= 1
        self.dimension = new_dim

        return self

    def periodic(self, direction=0):
        """Returns true if the spline object is periodic in the given parametric direction"""
        direction = check_direction(direction, self.pardim)

        return self.bases[direction].periodic > -1

    def force_rational(self):
        """  Force a rational representation of the object.

        The weights of a non-rational object will be set to 1.

        :return: self
        """
        if not self.rational:
            dim = self.dimension
            shape = self.controlpoints.shape
            self.controlpoints = np.insert(self.controlpoints, dim, np.ones(shape[:-1]), self.pardim)
            self.rational = 1

        return self

    def split(self, knots, direction=0):
        """  Split an object into two or more separate representations with C0
        continuity between them.

        :param knots: The splitting points
        :type knots: float or [float]
        :param direction: Parametric direction
        :type direction: int
        :return: The new objects
        :rtype: [SplineObject]
        """
        # for single-value input, wrap it into a list
        knots = ensure_listlike(knots)

        # error test input
        direction = check_direction(direction, self.pardim)

        p = self.order(direction)
        results = []
        splitting_obj = self.clone()
        bases = self.bases
        # insert knots to produce C{-1} at all splitting points
        for k in knots:
            continuity = bases[direction].continuity(k)
            if continuity == np.inf:
                continuity = p - 1
            splitting_obj.insert_knot([k] * (continuity + 1), direction)

        b = splitting_obj.bases[direction]
        if b.periodic > -1:
            mu = bisect_left(b.knots, knots[0])
            b.roll(mu)
            splitting_obj.controlpoints = np.roll(splitting_obj.controlpoints, -mu, direction)
            b.knots = b.knots[:-b.periodic-1]
            b.periodic = -1
            if len(knots) > 1:
                return splitting_obj.split(knots[1:], direction)
            else:
                return splitting_obj

        # search for the right subclass constructor, i.e. Volume, Surface or Curve
        spline_object = [c for c in SplineObject.__subclasses__() if c._intended_pardim == len(bases)]
        spline_object = spline_object[0]

        # everything is available now, just have to find the right index range
        # in the knot vector and controlpoints to store in each separate curve
        # piece
        last_cp_i = 0
        last_knot_i = 0

        bases = splitting_obj.bases
        b     = bases[direction]
        cp_slice = [slice(None, None, None)] * len(self.controlpoints.shape)
        for k in knots:
            if self.start(direction) < k < self.end(direction): # skip start/end points
                mu = bisect_left(b.knots, k)
                n_cp = mu - last_knot_i
                knot_slice          = slice(last_knot_i, mu+p, None)
                cp_slice[direction] = slice(last_cp_i,   last_cp_i+n_cp,  None)

                cp = splitting_obj.controlpoints[ tuple(cp_slice) ]
                bases[direction] = BSplineBasis(p, b.knots[knot_slice])

                args = bases + [cp, splitting_obj.rational]
                results.append(spline_object(*args, raw=True))

                last_knot_i = mu
                last_cp_i += n_cp

        # with n splitting points, we're getting n+1 pieces. Add the final one:
        knot_slice          = slice(last_knot_i, None, None)
        cp_slice[direction] = slice(last_cp_i,   None, None)
        bases[direction] = BSplineBasis(p, b.knots[knot_slice])
        cp = splitting_obj.controlpoints[ tuple(cp_slice) ]
        args = bases + [cp, splitting_obj.rational]
        results.append(spline_object(*args, raw=True))

        return results

    def make_periodic(self, continuity=None, direction=0):
        """  Make the spline object periodic in a given parametric direction.

        :param int continuity: The continuity along the boundary (default max).
        :param int direction: The direction to ensure continuity in.
        """

        direction = check_direction(direction, self.pardim)
        basis = self.bases[direction]
        if continuity is None:
            continuity = basis.order - 2
        if not -1 <= continuity <= basis.order - 2:
            raise ValueError('Illegal continuity for basis of order {}: {}'.format(
                continuity, basis.order
            ))
        if continuity == -1:
            raise ValueError(
                'Operation not supported. '
                'For discontinuous spline spaces, consider SplineObject.split().'
            )
        if basis.periodic >= 0:
            raise ValueError('Basis is already periodic')

        basis = basis.make_periodic(continuity)

        # Merge control points
        index_beg = [slice(None,None,None)] * (self.pardim + 1)
        index_end = [slice(None,None,None)] * (self.pardim + 1)
        cps = np.array(self.controlpoints)
        weights = np.linspace(0, 1, continuity + 1) if continuity > 0 else [0.5]
        for i, j, t in zip(range(continuity + 1), range(-continuity-1, 0), weights):
            # Weighted average between cps[..., i, ..., :] and cps[..., -c-1+i, ..., :]
            # The weights are chosen so that, for periodic c, the round trip
            # c.split().make_periodic() with suitable arguments produces an
            # object identical to c. (Mostly black magic.)
            index_beg[direction] = i
            index_end[direction] = j
            cps[tuple(index_beg)] = t * cps[tuple(index_beg)] + (1 - t) * cps[tuple(index_end)]

        # cps[..., :-(continuity+1), ..., :]
        index_beg[direction] = slice(None, -(continuity + 1), None)
        cps = cps[tuple(index_beg)]

        bases = list(self.bases)
        bases[direction] = basis
        args = bases + [cps] + [self.rational]

        # search for the right subclass constructor, i.e. Volume, Surface or Curve
        constructor = [c for c in SplineObject.__subclasses__() if c._intended_pardim == len(self.bases)]
        constructor = constructor[0]
        return constructor(*args, raw=True)

    @property
    def pardim(self):
        """  The number of parametric dimensions: 1 for curves, 2 for surfaces, 3
        for volumes, etc.
        """
        return len(self.controlpoints.shape)-1

    def clone(self):
        """Clone the object."""
        return copy.deepcopy(self)

    __call__ = evaluate

    def __len__(self):
        """Return the number of control points (basis functions) for the object."""
        n = 1
        for b in self.bases:
            n *= b.num_functions()
        return n

    def _unravel_flat_index(self, i):
        """Unravels a flat index i to multi-indexes.

        :param i: Flat index
        :type i: int or slice
        :rtype: Tuple of np.array
        :raises IndexError: If the index is out of bounds
        """
        # i is int => make sure we deal with negative i properly
        # i is slice => use i.indices to compute the actual indices
        total = len(self)
        if isinstance(i, int):
            indexes = [i] if i >= 0 else [total + i]
        else:
            indexes = list(range(*i.indices(total)))

        # Convert to multi-indexes
        try:
            unraveled = np.unravel_index(indexes, self.controlpoints.shape[:-1], order='F')
        except ValueError:
            raise IndexError

        return unraveled

    def __getitem__(self, i):
        """Get the control point at a given index.

        Indexing is in column-major order. Examples of supported indexing
        modes:

        .. code:: python

           # Flat indexing with an int
           obj[4]

           # Flat indexing from the end
           obj[-1]

           # Flat indexing with a slice
           obj[2:5]

           # Multi-indexing with ints, negative ints and slices
           obj[0,-1,:]

        :rtype: numpy.array
        """
        if isinstance(i, tuple):
            return self.controlpoints[i]

        unraveled = self._unravel_flat_index(i)

        # Singleton dimensions should be squeezed if the input was an int
        if isinstance(i, int):
            return self.controlpoints[unraveled][0]
        return self.controlpoints[unraveled]

    def __setitem__(self, i, cp):
        """Set the control points at given indices.

        This function supports the same indexing modes as
        :func:`SplineObject.__getitem__`

        :param int i: Index or indices
        :param numpy.array cp: New control point(s)
        """
        if isinstance(i, tuple):
            self.controlpoints[i] = cp
            return

        unraveled = self._unravel_flat_index(i)
        self.controlpoints[unraveled] = cp

    @property
    def shape(self):
        """The dimensions of the control point array."""
        return self.controlpoints.shape[:-1]

    def __iadd__(self, x):
        self.translate(x)
        return self

    def __isub__(self, x):
        self.translate(-np.array(x))  # can't do -x if x is a list, so we rewrap it here
        return self

    def __imul__(self, x):
        self.scale(x)
        return self

    def __itruediv__(self, x):
        self.scale(1.0 / x)
        return self

    __ifloordiv__ = __itruediv__  # integer division (should not distinguish)
    __idiv__ = __itruediv__  # python2 compatibility

    def __add__(self, x):
        new_obj = copy.deepcopy(self)
        new_obj += x
        return new_obj

    def __radd__(self, x):
        return self + x

    def __sub__(self, x):
        new_obj = copy.deepcopy(self)
        new_obj -= x
        return new_obj

    def __mul__(self, x):
        new_obj = copy.deepcopy(self)
        new_obj *= x
        return new_obj

    def __rmul__(self, x):
        return self * x

    def __div__(self, x):
        new_obj = copy.deepcopy(self)
        new_obj /= x
        return new_obj

    @classmethod
    def make_splines_compatible(cls, spline1, spline2):
        """Ensure that two splines are compatible.

        This will manipulate one or both to ensure that they are both rational
        or nonrational, and that they lie in the same physical space.

        :param SplineObject spline1: The first spline
        :param SplineObject spline2: The second spline
        """
        # make both rational (if needed)
        if spline1.rational:
            spline2.force_rational()
        elif spline2.rational:
            spline1.force_rational()

        # make both in the same geometric space
        if spline1.dimension > spline2.dimension:
            spline2.set_dimension(spline1.dimension)
        else:
            spline1.set_dimension(spline2.dimension)

    @classmethod
    def make_splines_identical(cls, spline1, spline2, direction=None):
        """Ensure that two splines have identical discretization.

        This will first make them compatible (see
        :func:`splipy.SplineObject.make_curves_compatible`), reparametrize them, and
        possibly raise the order and insert knots as required.

        :param SplineObject spline1: The first spline
        :param SplineObject spline2: The second spline
        :param int direction: The direction to make identical. If
            None, make all directions identical.
        """

        # make sure that rational/dimension is the same
        SplineObject.make_splines_compatible(spline1, spline2)

        # If all directions, just call the same method several times
        if direction is None:
            for i in range(spline1.pardim):
                cls.make_splines_identical(spline1, spline2, direction=i)
            return

        # From this point, assume we're running on a single direction
        i = check_direction(direction, spline1.pardim)

        # make both have knot vectors in domain (0,1)
        spline1.reparam(direction=i)
        spline2.reparam(direction=i)

        # settle on the lowest periodicity if different appear
        if spline1.bases[i].periodic < spline2.bases[i].periodic:
            spline2.lower_periodic(spline1.bases[i].periodic, i)
        elif spline2.bases[i].periodic < spline1.bases[i].periodic:
            spline1.lower_periodic(spline2.bases[i].periodic, i)

        # make sure both have the same order
        p1 = spline1.order(i)
        p2 = spline2.order(i)
        p = max(p1, p2)
        spline1.raise_order(p - p1, direction=i)
        spline2.raise_order(p - p2, direction=i)

        # make sure both have the same knot vectors
        knot1 = spline1.knots(direction=i)
        knot2 = spline2.knots(direction=i)
        b1    = spline1.bases[i]
        b2    = spline2.bases[i]

        inserts = []
        for k in knot1:
            c1 = b1.continuity(k)
            c2 = b2.continuity(k)
            if c2 > c1:
                m = min(c2-c1, p-1-c1) # c2=np.inf if knot does not exist
                inserts.extend([k] * m)
        spline2.insert_knot(inserts, direction=i)

        inserts = []
        for k in knot2:
            c1 = b1.continuity(k)
            c2 = b2.continuity(k)
            if c1 > c2:
                m = min(c1-c2, p-1-c2) # c1=np.inf if knot does not exist
                inserts.extend([k]*m)
        spline1.insert_knot(inserts, direction=i)



__doc__ = 'Implementation of various smoothing operations on a per-controlpoint level.'

from scipy import ndimage
import numpy as np

from . import check_direction


def smooth(obj, comp=None):
    """Smooth an object by setting the interior control points to the average of
    itself and all neighbours (e.g. 9 for surfaces, 27 for volumes). The edges
    are kept unchanged, and any rational weights are kept unchanged.

    :param obj: The object to smooth
    :type obj: :class:`splipy.SplineObject`
    :param comp: which component to smooth 0=x-component, 1=y-component, 2=z-component, None=all components
    :type comp: int or None
    """
    n = obj.shape
    if comp is not None:
        comp = check_direction(comp, len(obj))

    averaging_mask  = np.ones([3]*len(n)+[1])
    averaging_mask /= averaging_mask.size

    new_controlpoints = ndimage.convolve(obj.controlpoints, averaging_mask, mode='wrap')

    # build up the indexing for the domain 'interior'. This would be
    # controlpoints[1:-1, 1:-1 ,:]        for non-rational surface
    # controlpoints[ :  , 1:-1 ,:]        for surfaces which is periodic in 'u'
    # controlpoints[1:-1, 1:-1 ,:-1]      for rational surfaces
    # controlpoints[1:-1, 1:-1 , 1:-1, :] for non-rational volumes
    # controlpoints[ :  ,  :   ,  :  , :] for non-rational volumes which is periodic in all three parametric directions
    interior = []
    for pardim in range(len(n)):
        if obj.periodic(pardim):
            interior.append(slice(None,None,None))
        else:
            interior.append(slice(1,-1,None))
    if obj.rational:
        interior.append(slice(0,-1,None))
    elif comp is not None:
        interior.append(slice(comp,comp+1,None))
    else:
        interior.append(slice(None,None,None))

    interior = tuple(interior)
    obj.controlpoints[interior] =  new_controlpoints[interior]



__doc__ = 'Implementation of various curve utilities'

import numpy as np


def curve_length_parametrization(pts, normalize=False):
    """Calculate knots corresponding to a curvelength parametrization of a set of
    points.

    :param numpy.array pts: A set of points
    :param bool normalize: Whether to normalize the parametrization
    :return: The parametrization
    :rtype: [float]
    """
    knots = [0.0]
    for i in range(1, pts.shape[0]):
        knots.append(knots[-1] + np.linalg.norm(pts[i,:] - pts[i-1,:]))

    if normalize:
        length = knots[-1]
        knots = [k/length for k in knots]

    return knots


def get_curve_points(curve):
    """Evaluate the curve in all its knots.

    :param curve: The curve
    :type curve: :class:`splipy.Curve`
    :return: The curve points
    :type: numpy.array
    """
    return curve(curve.knots(0))



from itertools import chain, product

import numpy as np
from numpy import sqrt, pi, savetxt

from ..curve import Curve
from ..surface import Surface
from ..volume import Volume
from ..splineobject import SplineObject
from ..basis import BSplineBasis
from ..trimmedsurface import TrimmedSurface
from ..utils import flip_and_move_plane_geometry, rotate_local_x_axis
from .. import surface_factory, curve_factory, state

from .master import MasterIO


class G2(MasterIO):

    def read_next_non_whitespace(self):
        line = next(self.fstream).strip()
        while not line:
            line = next(self.fstream).strip()
        return line

    def circle(self):
        dim   = int(     self.read_next_non_whitespace().strip())
        r     = float(   next(self.fstream).strip())
        center= np.array(next(self.fstream).split(), dtype=float)
        normal= np.array(next(self.fstream).split(), dtype=float)
        xaxis = np.array(next(self.fstream).split(), dtype=float)
        param = np.array(next(self.fstream).split(), dtype=float)
        reverse =        next(self.fstream).strip() != '0'

        result = curve_factory.circle(r=r, center=center, normal=normal, xaxis=xaxis)
        result.reparam(param)
        if reverse:
            result.reverse()
        return result

    def ellipse(self):
        dim   = int(     self.read_next_non_whitespace().strip())
        r1    = float(   next(self.fstream).strip())
        r2    = float(   next(self.fstream).strip())
        center= np.array(next(self.fstream).split(), dtype=float)
        normal= np.array(next(self.fstream).split(), dtype=float)
        xaxis = np.array(next(self.fstream).split(), dtype=float)
        param = np.array(next(self.fstream).split(), dtype=float)
        reverse =        next(self.fstream).strip() != '0'

        result = curve_factory.ellipse(r1=r1, r2=r2, center=center, normal=normal, xaxis=xaxis)
        result.reparam(param)
        if reverse:
            result.reverse()
        return result

    def line(self):
        dim      = int(     self.read_next_non_whitespace().strip())
        start    = np.array(next(self.fstream).split(), dtype=float)
        direction= np.array(next(self.fstream).split(), dtype=float)
        finite   =          next(self.fstream).strip() != '0'
        param    = np.array(next(self.fstream).split(), dtype=float)
        reverse  =          next(self.fstream).strip() != '0'
        d = np.array(direction)
        s = np.array(start)
        # d /= np.linalg.norm(d)
        if not finite:
            param = [-state.unlimited, +state.unlimited]

        result = curve_factory.line(s+d*param[0], s+d*param[1])
        if reverse:
            result.reverse()
        return result


#   def cone(self):
#       dim      = int(     self.read_next_non_whitespace().strip())
#       r        = float(   next(self.fstream).strip())
#       center   = np.array(next(self.fstream).split(' '), dtype=float)
#       z_axis   = np.array(next(self.fstream).split(' '), dtype=float)
#       x_axis   = np.array(next(self.fstream).split(' '), dtype=float)
#       angle    = float(   next(self.fstream).strip())
#       finite   =          next(self.fstream).strip() != '0'
#       param_u  = np.array(next(self.fstream).split(' '), dtype=float)
#       if finite:
#           param_v=np.array(next(self.fstream).split(' '), dtype=float)


    def cylinder(self):
        dim      = int(     self.read_next_non_whitespace().strip())
        r        = float(   next(self.fstream).strip())
        center   = np.array(next(self.fstream).split(), dtype=float)
        z_axis   = np.array(next(self.fstream).split(), dtype=float)
        x_axis   = np.array(next(self.fstream).split(), dtype=float)
        finite   =          next(self.fstream).strip() != '0'
        param_u  = np.array(next(self.fstream).split(), dtype=float)
        if finite:
            param_v=np.array(next(self.fstream).split(), dtype=float)
        else:
            param_v=[-state.unlimited, state.unlimited]
        swap     =          next(self.fstream).strip() != '0'

        center = center + z_axis*param_v[0]
        h      = param_v[1] - param_v[0]
        result = surface_factory.cylinder(r=r, center=center, xaxis=x_axis, axis=z_axis, h=h)
        result.reparam(param_u, param_v)
        if swap:
            result.swap()
        return result

    def disc(self):
        dim      = int(     self.read_next_non_whitespace().strip())
        center   = np.array(next(self.fstream).split(), dtype=float)
        r        = float(   next(self.fstream).strip())
        z_axis   = np.array(next(self.fstream).split(), dtype=float)
        x_axis   = np.array(next(self.fstream).split(), dtype=float)
        degen    =          next(self.fstream).strip() != '0'
        angles   = [float(  next(self.fstream).strip()) for i in range(4)]
        param_u  = np.array(next(self.fstream).split(), dtype=float)
        param_v  = np.array(next(self.fstream).split(), dtype=float)
        swap     =          next(self.fstream).strip() != '0'

        if degen:
            result = surface_factory.disc(r=r, center=center, xaxis=x_axis, normal=z_axis, type='radial')
        else:
            if not(np.allclose(np.diff(angles), pi/2, atol=1e-10)):
                raise RuntimeError('Unknown square parametrization of disc elementary surface')
            result = surface_factory.disc(r=r, center=center, xaxis=x_axis, normal=z_axis, type='square')
        result.reparam(param_u, param_v)
        if swap:
            result.swap()
        return result

    def plane(self):
        dim        = int(     self.read_next_non_whitespace().strip())
        center     = np.array(next(self.fstream).split(), dtype=float)
        normal     = np.array(next(self.fstream).split(), dtype=float)
        x_axis     = np.array(next(self.fstream).split(), dtype=float)
        finite     =          next(self.fstream).strip() != '0'
        if finite:
            param_u= np.array(next(self.fstream).split(), dtype=float)
            param_v= np.array(next(self.fstream).split(), dtype=float)
        else:
            param_u= [-state.unlimited, +state.unlimited]
            param_v= [-state.unlimited, +state.unlimited]
        swap       =          next(self.fstream).strip() != '0'

        result = Surface() * [param_u[1]-param_u[0], param_v[1]-param_v[0]] + [param_u[0],param_v[0]]
        result.rotate(rotate_local_x_axis(x_axis, normal))
        result = flip_and_move_plane_geometry(result,center,normal)
        result.reparam(param_u, param_v)
        if(swap):
            result.swap()
        return result

    def torus(self):
        dim      = int(     self.read_next_non_whitespace().strip())
        r2       = float(   next(self.fstream).strip())
        r1       = float(   next(self.fstream).strip())
        center   = np.array(next(self.fstream).split(), dtype=float)
        z_axis   = np.array(next(self.fstream).split(), dtype=float)
        x_axis   = np.array(next(self.fstream).split(), dtype=float)
        select_out=         next(self.fstream).strip() != '0' # I have no idea what this does :(
        param_u  = np.array(next(self.fstream).split(), dtype=float)
        param_v  = np.array(next(self.fstream).split(), dtype=float)
        swap     =          next(self.fstream).strip() != '0'

        result = surface_factory.torus(minor_r=r1, major_r=r2, center=center, normal=z_axis, xaxis=x_axis)
        result.reparam(param_u, param_v)
        if(swap):
            result.swap()
        return result

    def sphere(self):
        dim      = int(     self.read_next_non_whitespace().strip())
        r        = float(   next(self.fstream).strip())
        center   = np.array(next(self.fstream).split(), dtype=float)
        z_axis   = np.array(next(self.fstream).split(), dtype=float)
        x_axis   = np.array(next(self.fstream).split(), dtype=float)
        param_u  = np.array(next(self.fstream).split(), dtype=float)
        param_v  = np.array(next(self.fstream).split(), dtype=float)
        swap     =          next(self.fstream).strip() != '0'

        result = surface_factory.sphere(r=r, center=center, xaxis=x_axis, zaxis=z_axis).swap()
        if swap:
            result.swap()
        result.reparam(param_u, param_v)
        return result

    def splines(self, pardim):
        cls    = G2.classes[pardim-1]

        _, rational = self.read_next_non_whitespace().strip().split()
        rational = bool(int(rational))

        bases = [self.read_basis() for _ in range(pardim)]
        ncps = 1
        for b in bases:
            ncps *= b.num_functions()

        cps = [tuple(map(float, next(self.fstream).split()))
               for _ in range(ncps)]

        args = bases + [cps, rational]
        return cls(*args)

    def surface_of_linear_extrusion(self):
        dim      = int(      self.read_next_non_whitespace().strip())
        crv      = self.splines(1)
        normal   = np.array(self.read_next_non_whitespace().split(), dtype=float)
        finite   =          next(self.fstream).strip() != '0'
        param_u  = np.array(next(self.fstream).split(), dtype=float)
        if finite:
            param_v=np.array(next(self.fstream).split(), dtype=float)
        else:
            param_v=[-state.unlimited, +state.unlimited]
        swap     =          next(self.fstream).strip() != '0'

        result = surface_factory.extrude(crv + normal*param_v[0], normal*(param_v[1]-param_v[0]))
        result.reparam(param_u, param_v)

        if swap:
            result.swap()
        return result


    def bounded_surface(self):
        objtype = int( next(self.fstream).strip() )

        # create the underlying surface which all trimming curves are to be applied
        if objtype in G2.g2_generators:
            constructor = getattr(self, G2.g2_generators[objtype].__name__)
            surface = constructor()
        elif objtype == 200:
            surface = self.splines(2)
        else:
            raise IOError('Unsopported trimmed surface or malformed input file')

        # for all trimming loops
        numb_loops = int( self.read_next_non_whitespace() )
        all_loops  = []
        for i in range(numb_loops):

            # for all cuve pieces of that loop
            numb_crvs, space_epsilon = next(self.fstream).split()
            state.parametric_absolute_tolerance = float(space_epsilon)
            one_loop = []
            for j in range(int(numb_crvs)):

                # read a physical and parametric representation of the same curve
                _, parameter_curve_type, space_curve_type = map(int, self.read_next_non_whitespace().split())
                two_curves = []
                for crv_type in [parameter_curve_type, space_curve_type]:
                    if crv_type in G2.g2_generators:
                        constructor = getattr(self, G2.g2_generators[crv_type].__name__)
                        crv = constructor()
                    elif crv_type == 100:
                        crv = self.splines(1)
                    else:
                        raise IOError('Unsopported trimming curve or malformed input file')
                    two_curves.append(crv)

                # only keep the parametric version (re-generate physical one if we need it)
                one_loop.append(two_curves[0])
                self.trimming_curves.append(two_curves[1])
            all_loops.append(one_loop)

        return TrimmedSurface(surface.bases[0], surface.bases[1], surface.controlpoints, surface.rational, all_loops, raw=True)

    g2_type = [100, 200, 700] # curve, surface, volume identifiers
    classes = [Curve, Surface, Volume]
    g2_generators = {120:line, 130:circle, 140:ellipse,
                     260:cylinder, 292:disc, 270:sphere, 290:torus, 250:plane,
                     210:bounded_surface, 261:surface_of_linear_extrusion} #, 280:cone

    def __init__(self, filename):
        if filename[-3:] != '.g2':
            filename += '.g2'
        self.filename = filename
        self.trimming_curves = []

    def __enter__(self):
        return self

    def write(self, obj):
        if not hasattr(self, 'fstream'):
            self.onlywrite = True
            self.fstream   = open(self.filename, 'w')
        if not self.onlywrite:
            raise IOError('Could not write to file %s' % (self.filename))

        """Write the object in GoTools format. """
        if isinstance(obj[0], SplineObject): # input SplineModel or list
            for o in obj:
                self.write(o)
            return

        for i in range(obj.pardim):
            if obj.periodic(i):
                obj = obj.split(obj.start(i), i)

        self.fstream.write('{} 1 0 0\n'.format(G2.g2_type[obj.pardim-1]))
        self.fstream.write('{} {}\n'.format(obj.dimension, int(obj.rational)))
        for b in obj.bases:
            self.fstream.write('%i %i\n' % (len(b.knots) - b.order, b.order))
            self.fstream.write(' '.join('%.16g' % k for k in b.knots))
            self.fstream.write('\n')

        savetxt(self.fstream, obj.controlpoints.reshape(-1, obj.dimension + obj.rational, order='F'),
                fmt='%.16g', delimiter=' ', newline='\n')

    def read(self):
        if not hasattr(self, 'fstream'):
            self.onlywrite = False
            self.fstream   = open(self.filename, 'r')

        if self.onlywrite:
            raise IOError('Could not read from file %s' % (self.filename))

        result = []

        for line in self.fstream:
            line = line.strip()
            if not line:
                continue

            # read object type
            objtype, major, minor, patch = map(int, line.split())
            if (major, minor, patch) != (1, 0, 0):
                raise IOError('Unknown G2 format')

            # if obj type is in factory methods (cicle, torus etc), create it now
            if objtype in G2.g2_generators:
                constructor = getattr(self, G2.g2_generators[objtype].__name__)
                result.append( constructor() )
                continue

            # for "normal" splines (Curves, Surfaces, Volumes) create it now
            pardim = [i for i in range(len(G2.g2_type)) if G2.g2_type[i] == objtype]
            if not pardim:
                raise IOError('Unknown G2 object type {}'.format(objtype))
            pardim = pardim[0] + 1
            result.append(self.splines(pardim))

        return result

    def read_basis(self):
        ncps, order = map(int, next(self.fstream).split())
        kts = list(map(float, next(self.fstream).split()))
        return BSplineBasis(order, kts, -1)

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(self, 'fstream'):
            self.fstream.close()



import xml.etree.ElementTree as etree
from xml.dom import minidom
import re

import numpy as np

from ..curve import Curve
from ..surface import Surface
from ..splineobject import SplineObject
from ..basis import BSplineBasis
from .. import curve_factory, state

from .master import MasterIO


def read_number_and_unit(mystring):
    unit = ''
    try:
        for i in range(1, len(mystring)+1):
            number=float(mystring[:i])
    except ValueError:
        unit=mystring[i-1:]
    return (number, unit)


def bezier_representation(curve):
    """ Compute a Bezier representation of a given spline curve. The input
        curve must be of order less than or equal to 4. The bezier
        representation is of order 4, and maximal knot multiplicity, i.e.
        consist of a C0-discretization with knot multiplicity 3.

        :param curve  : Spline curve
        :type  curve  : Curve
        :returns      : Bezier curve
        :rtype        : Curve
    """
    # error test input. Another way would be to approximate higher-order curves. Consider looking at Curve.rebuild()
    if curve.order(0) > 4 or curve.rational:
        raise RuntimeError('Bezier representation, only supported for non-rational curves of order 4 or less')

    bezier = curve.clone()
    bezier.raise_order(4-curve.order(0)) # make sure curve is cubic

    # make it non-periodic
    if bezier.periodic():
        bezier = bezier.split(bezier.start(0))

    # make sure it is C0 everywhere
    for k in bezier.knots(0):
        bezier.insert_knot( [k]*bezier.continuity(k) )

    return bezier


class SVG(MasterIO):

    namespace = '{http://www.w3.org/2000/svg}'

    def __init__(self, filename, width=1000, height=1000, margin=0.05):
        """  Constructor
          :param filename: Filename to write results to
          :type  filename: String
          :param width   : Maximum image width in pixels
          :type  width   : Int
          :param height  : Maximum image height in pixels
          :type  height  : Int
          :param margin  : White-space around all edges of image, given in percentage of total size (default 5%)
          :type  margin  : Float
        """
        if filename[-4:] != '.svg':
            filename += '.svg'
        self.filename = filename

        self.width  = width
        self.height = height
        self.margin = margin

        self.all_objects = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # in case something goes wrong, print error and abandon writing process
        if exc_type is not None:
            print(exc_type, exc_value, traceback)
            return False

        # compute the bounding box for all geometries
        boundingbox = [np.inf, np.inf, -np.inf, -np.inf]
        for entry in self.all_objects:
            bb = entry.bounding_box()
            boundingbox[0] = min(boundingbox[0], bb[0][0])
            boundingbox[1] = min(boundingbox[1], bb[1][0])
            boundingbox[2] = max(boundingbox[2], bb[0][1])
            boundingbox[3] = max(boundingbox[3], bb[1][1])

        # compute scaling factors by keeping aspect ratio, and never exceed width or height size (including margins)
        geometryRatio = float(boundingbox[3]-boundingbox[1])/(boundingbox[2]-boundingbox[0])
        imageRatio    = 1.0*self.height / self.width
        if geometryRatio > imageRatio: # scale by y-coordinate
            marginPixels = self.height*self.margin
            self.scale   = self.height*(1-2*self.margin) / (boundingbox[3]-boundingbox[1])
            self.width   = self.height/geometryRatio + 2*marginPixels
        else:                          # scale by x-coordinate
            marginPixels = self.width*self.margin
            self.scale   = self.width*(1-2*self.margin) / (boundingbox[2]-boundingbox[0])
            self.height  = self.width*geometryRatio + 2*marginPixels
        self.center      = [boundingbox[0], boundingbox[1]]
        self.offset      = [marginPixels, marginPixels]

        # create xml root tag
        self.xmlRoot = etree.Element('svg',  {'xmlns':'http://www.w3.org/2000/svg',
                                              'version':'1.1',
                                              'width':str(self.width),
                                              'height':str(self.height)})

        # populate tree with all curves and surfaces in entities
        for entry in self.all_objects:
            if isinstance(entry, Curve):
                self.write_curve(self.xmlRoot, entry)
            elif isinstance(entry, Surface):
                self.write_surface(entry)

        # if no objects are stored, then we've most likely only called read()
        if len(self.all_objects) > 0:
            rough_string = etree.tostring(self.xmlRoot) # entire xml-file on one line
            reparsed     = minidom.parseString(rough_string)
            result       = reparsed.toprettyxml(indent="  ") # adds newline and inline
            f = open(self.filename, 'w')
            f.write(result)

    def write_curve(self, xmlNode, curve, fill='none', stroke='#000000', width=2):
        """ Writes a Curve to the xml tree. This will draw a single curve

            :param xmlNode: Node in xml tree
            :type  xmlNode: Etree.element
            :param curve  : The spline curve to write
            :type  curve  : Curve
            :param fill   : Fill color in hex or none
            :type  fill   : String
            :param stroke : Line color written in hex, i.e. '#000000'
            :type  stroke : String
            :param width  : Line width, measured in pixels
            :type  width  : Int
            :returns: None
            :rtype  : NoneType
        """
        curveNode = etree.SubElement(xmlNode, 'path')
        curveNode.attrib['style'] = 'fill:%s;stroke:%s;stroke-width:%dpx;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1' %(fill,stroke,width)
        bezier  = bezier_representation(curve)
        bezier -= self.center
        bezier *= self.scale
        bezier += self.offset
        pathString = 'M %f,%f C ' % (bezier[0][0],  self.height + 2*self.margin - bezier[0][1])

        for i in range(1,len(bezier)):
            pathString += '%f,%f ' % (bezier[i][0], self.height + 2*self.margin - bezier[i][1])

        curveNode.attrib['d'] = pathString

    def write_surface(self, surface, fill='#ffcc99'):
        """ Writes a Surface to the xml tree. This will draw the surface along with all knot lines

            :param surface: The spline surface to write
            :type  surface: Surface
            :param fill   : Surface color written in hex, i.e. '#ffcc99'
            :type  fill   : String
            :returns: None
            :rtype  : NoneType
        """
        # fetch boundary curves and create a connected, oriented bezier loop from it
        bndry_curves = surface.edges()
        bndry_curves[0].reverse()
        bndry_curves[3].reverse()
        boundary      = bndry_curves[0]
        boundary.append(bndry_curves[2])
        boundary.append(bndry_curves[1])
        boundary.append(bndry_curves[3])

        # fetch all meshlines (i.e. elements, also known as knot spans)
        knot = surface.knots()
        knotlines = []
        for k in knot[0][1:-1]:
            knotlines.append(surface.const_par_curve(k, 0))
        for k in knot[1][1:-1]:
            knotlines.append(surface.const_par_curve(k, 1))

        # create a group node for all elements corresponding to this surface patch
        groupNode = etree.SubElement(self.xmlRoot, 'g')

        # fill interior with a peach color
        self.write_curve(groupNode, boundary, fill, width=2)

        # draw all meshlines
        for meshline in knotlines:
            self.write_curve(groupNode, meshline, width=1)


    def write(self, obj):
        """ Writes a list of planar curves and surfaces to vector graphics SVG file.
            The image will never be stretched, and the geometry will keep width/height ratio
            of original geometry, regardless of provided width/height ratio from arguments.

            :param obj: Primitives to write
            :type  obj: List of Curves and/or Surfaces
            :returns: None
            :rtype  : NoneType
        """
        # actually this is a dummy method. It will collect all geometries provided
        # and ton't actually write them to file until __exit__ is called

        if isinstance(obj[0], SplineObject): # input SplineModel or list
            for o in obj:
                self.write(o)
            return

        if obj.dimension != 2:
            raise RuntimeError('SVG files only applicable for 2D geometries')

        # have to clone stuff we put here, in case they change on the outside
        self.all_objects.append( obj.clone() )

    def read(self):
        tree = etree.parse(self.filename)
        root = tree.getroot()
        parent_map = dict((c, p) for p in tree.iter() for c in p)
        if 'width' in root.attrib:
            self.width,_  = read_number_and_unit(root.attrib['width'])
            self.height,_ = read_number_and_unit(root.attrib['height'])
        result = []
        for path in root.iter(SVG.namespace + 'path'):
            crvs = self.curves_from_path(path.attrib['d'])
            parent = path
            while parent != root:
                if 'transform' in parent.attrib:
                    for crv in crvs:
                        self.transform(crv, parent.attrib['transform'])
                parent = parent_map[parent]

            # invert y-values since these are image coordinates
            for crv in crvs:
                crv *= [1,-1]
                crv += [0, self.height]
                result.append(crv)
        return result

    def transform(self, curve, operation):
        # intended input operation string: 'translate(-10,-20) scale(2) rotate(45) translate(5,10)'
        all_operations = re.findall(r'[^\)]*\)', operation.lower())
        all_operations.reverse()
        for one_operation in all_operations:
            parts = re.search(r'([a-z]*)\w*\((.*)\)', one_operation.strip())
            func = parts.group(1)
            args = [float(d) for d in parts.group(2).split(',')]
            if func == 'translate':
                if len(args)==1:
                    args.append(0)
                curve += args
            elif func == 'scale':
                if len(args)==1:
                    args.append(args[0])
                curve *= args
            elif func == 'rotate':
                curve.rotate(-args[0]/360*2*np.pi)
            elif func == 'matrix':
                M = np.array([[args[0], args[2], args[4]],
                               [args[1], args[3], args[5]],
                               [      0,       0,      1]])
                n = len(curve)
                if not curve.rational:
                    cp = np.ones((n, 3))  # pad with weights=1
                    cp[:, :-1] = np.reshape(curve.controlpoints, (n, 2))
                    cp = cp @ M.T
                    curve.controlpoints = np.reshape(np.array(cp[:,:-1]), curve.controlpoints.shape)
                else:
                    cp = np.reshape(curve.controlpoints, (n, 3))
                    cp = cp @ M.T
                    curve.controlpoints = np.reshape(np.array(cp), curve.controlpoints.shape)

    def curves_from_path(self, path):
        # see https://www.w3schools.com/graphics/svg_path.asp for documentation
        # and also https://www.w3.org/TR/SVG/paths.html

        # figure out the largest polynomial order of this path
        if re.search('[cCsS]', path):
            order = 4
        elif re.search('[qQtTaA]', path):
            order = 3
        else:
            order = 2
        last_curve = None
        result = []

        # each 'piece' is an operator (M,C,Q,L etc) and accomponying list of argument points
        for piece in re.findall('[a-zA-Z][^a-zA-Z]*', path):

            # if not single-letter command (i.e. 'z')
            if len(piece)>1:
                # points is a (string-)list of (x,y)-coordinates for the given operator
                points = re.findall(r'-?\d+\.?\d*', piece[1:])

                if piece[0].lower() != 'a' and piece[0].lower() != 'v' and piece[0].lower() != 'h':
                    # convert string-list to a list of numpy arrays (of size 2)
                    np_pts = np.reshape(np.array(points).astype('float'), (int(len(points)/2),2))

            if piece[0] == 'm' or piece[0] == 'M':
                # I really hope it always start with a move command (think it does)
                startpoint = np_pts[0]
                if len(np_pts) > 1:
                    if piece[0] == 'M':
                        knot = [0] + list(range(len(np_pts))) + [len(np_pts)-1]
                        curve_piece = Curve(BSplineBasis(2, knot), np_pts)
                    elif piece[0] == 'm':
                        knot = [0] + list(range(len(np_pts))) + [len(np_pts)-1]
                        controlpoints = [startpoint]
                        for cp in np_pts[1:]:
                            controlpoints.append(cp + controlpoints[-1])
                        curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
                else:
                    continue
            elif piece[0] == 'c':
                # cubic spline, relatively positioned
                controlpoints = [startpoint]
                knot = list(range(int(len(np_pts)/3)+1)) * 3
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    startpoint = controlpoints[int((len(controlpoints)-1)/3)*3]
                    controlpoints.append(cp + startpoint)
                curve_piece = Curve(BSplineBasis(4, knot), controlpoints)
            elif piece[0] == 'C':
                # cubic spline, absolute position
                controlpoints = [startpoint]
                knot = list(range(int(len(np_pts)/3)+1)) * 3
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    controlpoints.append(cp)
                curve_piece = Curve(BSplineBasis(4, knot), controlpoints)
            elif piece[0] == 's':
                # smooth cubic spline, relative position
                controlpoints = [startpoint]
                knot = list(range(int(len(np_pts)/2)+1)) * 3
                knot += [knot[0], knot[-1]]
                knot.sort()
                x0  = np.array(last_curve[-1])
                xn1 = np.array(last_curve[-2])
                controlpoints.append(2*x0 -xn1)
                startpoint = controlpoints[-1]
                for i, cp in enumerate(np_pts):
                    if i % 2 == 0 and i>0:
                        startpoint = controlpoints[-1]
                        controlpoints.append(2*controlpoints[-1] - controlpoints[-2])
                    controlpoints.append(cp + startpoint)
                curve_piece = Curve(BSplineBasis(4, knot), controlpoints)
            elif piece[0] == 'S':
                # smooth cubic spline, absolute position
                controlpoints = [startpoint]
                knot = list(range(int(len(np_pts)/2)+1)) * 3
                knot += [knot[0], knot[-1]]
                knot.sort()
                x0  = np.array(last_curve[-1])
                xn1 = np.array(last_curve[-2])
                controlpoints.append(2*x0 -xn1)
                for i,cp in enumerate(np_pts):
                    if i % 2 == 0 and i>0:
                        controlpoints.append(2*controlpoints[-1] - controlpoints[-2])
                    controlpoints.append(cp)
                curve_piece = Curve(BSplineBasis(4, knot), controlpoints)
            elif piece[0] == 'q':
                # quadratic spline, relatively positioned
                controlpoints = [startpoint]
                knot = list(range(int(len(np_pts)/2)+1)) * 2
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    startpoint = controlpoints[int((len(controlpoints)-1)/2)*2]
                    controlpoints.append(cp + startpoint)
                curve_piece = Curve(BSplineBasis(3, knot), controlpoints)
            elif piece[0] == 'Q':
                # quadratic spline, absolute position
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)/2+1)) * 2
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    controlpoints.append(cp)
                curve_piece = Curve(BSplineBasis(3, knot), controlpoints)
            elif piece[0] == 'l':
                # linear spline, relatively positioned
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)+1))
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    startpoint = controlpoints[-1]
                    controlpoints.append(cp + startpoint)
                curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
            elif piece[0] == 'L':
                # linear spline, absolute position
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)+1))
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    controlpoints.append(cp)
                curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
            elif piece[0] == 'h':
                # horizontal piece, relatively positioned
                np_pts = np.array(points).astype('float')
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)+1))
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    startpoint = controlpoints[-1]
                    controlpoints.append(np.array([cp, 0]) + startpoint)
                curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
            elif piece[0] == 'H':
                # horizontal piece, absolute position
                np_pts = np.array(points).astype('float')
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)+1))
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    controlpoints.append([cp, startpoint[1]])
                curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
            elif piece[0] == 'v':
                # vertical piece, relatively positioned
                np_pts = np.array(points).astype('float')
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)+1))
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    startpoint = controlpoints[-1]
                    controlpoints.append(np.array([0, cp]) + startpoint)
                curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
            elif piece[0] == 'V':
                # vertical piece, absolute position
                np_pts = np.array(points).astype('float')
                controlpoints = [startpoint]
                knot = list(range(len(np_pts)+1))
                knot += [knot[0], knot[-1]]
                knot.sort()
                for cp in np_pts:
                    controlpoints.append([startpoint[0], cp])
                curve_piece = Curve(BSplineBasis(2, knot), controlpoints)
            elif piece[0] == 'A' or piece[0] == 'a':
                np_pts = np.reshape(np.array(points).astype('float'), (int(len(points))))
                rx              = float(points[0])
                ry              = float(points[1])
                x_axis_rotation = float(points[2])
                large_arc_flag  = (points[3] != '0')
                sweep_flag      = (points[4] != '0')
                xend            = np.array([float(points[5]), float(points[6]) ])
                if piece[0] == 'a':
                    xend += startpoint

                R = np.array([[ np.cos(x_axis_rotation), np.sin(x_axis_rotation)],
                              [-np.sin(x_axis_rotation), np.cos(x_axis_rotation)]])
                xp = np.linalg.solve(R, (startpoint - xend)/2)
                if sweep_flag == large_arc_flag:
                    cprime = -(np.sqrt(abs(rx**2*ry**2 - rx**2*xp[1]**2 - ry**2*xp[0]**2) /
                                       (rx**2*xp[1]**2 + ry**2*xp[0]**2)) *
                                       np.array([rx*xp[1]/ry, -ry*xp[0]/rx]))
                else:
                    cprime = +(np.sqrt(abs(rx**2*ry**2 - rx**2*xp[1]**2 - ry**2*xp[0]**2) /
                                       (rx**2*xp[1]**2 + ry**2*xp[0]**2)) *
                                       np.array([rx*xp[1]/ry, -ry*xp[0]/rx]))
                center = np.linalg.solve(R.T, cprime) + (startpoint+xend)/2
                def arccos(vec1, vec2):
                    return (np.sign(vec1[0]*vec2[1] - vec1[1]*vec2[0]) *
                            np.arccos(vec1.dot(vec2)/np.linalg.norm(vec1)/np.linalg.norm(vec2)))
                tmp1 = np.divide( xp - cprime, [rx,ry])
                tmp2 = np.divide(-xp - cprime, [rx,ry])
                theta1 = arccos(np.array([1,0]), tmp1)
                delta_t= arccos(tmp1, tmp2) % (2*np.pi)
                if not sweep_flag and delta_t > 0:
                    delta_t -= 2*np.pi
                elif sweep_flag and delta_t < 0:
                    delta_t += 2*np.pi
                curve_piece = (curve_factory.circle_segment(delta_t)*[rx,ry]).rotate(theta1) + center
                # curve_piece = curve_factory.circle_segment(delta_t)

            elif piece[0] == 'z' or piece[0] == 'Z':
                # periodic curve
                # curve_piece = Curve(BSplineBasis(2), [startpoint, last_curve[0]])
                # curve_piece.reparam([0, curve_piece.length()])
                # last_curve.append(curve_piece).make_periodic(0)
                last_curve.make_periodic(0)
                result.append(last_curve)
                last_curve = None
                continue
            else:
                raise RuntimeError('Unknown path parameter:' + piece)

            if(curve_piece.length()>state.controlpoint_absolute_tolerance):
                curve_piece.reparam([0, curve_piece.length()])
                if last_curve is None:
                    last_curve = curve_piece
                else:
                    last_curve.append(curve_piece)
            startpoint = last_curve[-1,:2] # disregard rational weight (if any)

        if last_curve is not None:
            result.append(last_curve)
        return result



# -*- coding: utf-8 -*-

from bisect import bisect_left
from itertools import chain

import numpy as np

from .basis import BSplineBasis
from .curve import Curve
from .splineobject import SplineObject, evaluate
from .utils import is_singleton, ensure_listlike, check_direction, sections

__all__ = ['Surface']


class Surface(SplineObject):
    """Surface()

    Represents a surface: an object with a two-dimensional parameter space."""

    _intended_pardim = 2

    def __init__(self, basis1=None, basis2=None, controlpoints=None, rational=False, **kwargs):
        """  Construct a surface with the given basis and control points.

        The default is to create a linear one-element mapping from and to the
        unit square.

        :param BSplineBasis basis1: The basis of the first parameter direction
        :param BSplineBasis basis2: The basis of the second parameter direction
        :param array-like controlpoints: An *n1* × *n2* × *d* matrix of control points
        :param bool rational: Whether the surface is rational (in which case the
            control points are interpreted as pre-multiplied with the weight,
            which is the last coordinate)
        """
        super(Surface, self).__init__([basis1, basis2], controlpoints, rational, **kwargs)

    def normal(self, u, v, above=(True,True), tensor=True):
        """  Evaluate the normal of the surface at given parametric values.

        This is equal to the cross-product between tangents. The return value
        is normalized.

        If *tensor* is true, evaluation will take place on a tensor product
        grid, i.e. it will return an *n1* × *n2* × ... × *dim* array, where
        *ni* is the number of evaluation points in direction *i*, and *dim* is
        the physical dimension of the object.

        If *tensor* is false, there must be an equal number *n* of evaluation
        points in all directions, and the return value will be an *n* × *dim*
        array.

        :param u: Parametric coordinate(s) in the first direction
        :type u: float or [float]
        :param v: Parametric coordinate(s) in the second direction
        :type v: float or [float]
        :param (bool) above: Evaluation in the limit from above
        :param tensor: Whether to evaluate on a tensor product grid
        :type tensor: bool
        :return: Normal array *X[i,j,k]* of component *xk* evaluated at *(u[i], v[j])*
        :rtype: numpy.array
        :raises RuntimeError: If the physical dimension is not 2 or 3
        """
        if not tensor and len(u) != len(v):
            raise ValueError('Parametes must have same length')

        if self.dimension == 2:
            try:
                shape = (len(u), len(v), 3) if tensor else (len(u), 3)
                result = np.zeros(shape)
                result[..., 2] = 1
                return result
            except TypeError:  # single valued input u, fails on len(u)
                return np.array([0, 0, 1])
        elif self.dimension == 3:
            # fetch the tangent vectors
            (du, dv) = self.tangent(u, v, above=above, tensor=tensor)

            # compute normals
            normals = np.cross(du,dv)

            # normalize output
            if len(du.shape) == 1:
                return normals / np.linalg.norm(normals)
            magnitude = np.linalg.norm( normals, axis=-1)
            magnitude = magnitude.reshape(magnitude.shape + (1,))
            return normals / magnitude
        else:
            raise RuntimeError('Normal evaluation only defined for 2D and 3D geometries')


    def derivative(self, u, v, d=(1,1), above=True, tensor=True):
        """  Evaluate the derivative of the surface at the given parametric values.

        This function returns an *n* × *m* x *dim* array, where *n* is the number of
        evaluation points in u, *m* is the number of evaluation points in v, and
        *dim* is the physical dimension of the curve.

        If there is only one evaluation point, a vector of length *dim* is
        returned instead.

        :param u: Parametric coordinate(s) in the first direction
        :type u: float or [float]
        :param v: Parametric coordinate(s) in the second direction
        :type v: float or [float]
        :param int d: Number of derivatives to compute in each direction
        :type d: [int]
        :param (bool) above: Evaluation in the limit from above
        :param tensor: Whether to evaluate on a tensor product grid
        :type tensor: bool
        :return: Derivative array *X[i,j,k]* of component *xk* evaluated at *(u[i], v[j])*
        :rtype: numpy.array
        """

        squeeze = all(is_singleton(t) for t in [u,v])
        derivs = ensure_listlike(d, self.pardim)
        if not self.rational or np.sum(derivs) < 2 or np.sum(derivs) > 3:
            return super(Surface, self).derivative(u,v, d=derivs, above=above, tensor=tensor)

        u = ensure_listlike(u)
        v = ensure_listlike(v)
        result = np.zeros((len(u), len(v), self.dimension))
        # dNus = [self.bases[0].evaluate(u, d, above) for d in range(derivs[0]+1)]
        # dNvs = [self.bases[1].evaluate(v, d, above) for d in range(derivs[1]+1)]
        dNus = [self.bases[0].evaluate(u, d, above) for d in range(np.sum(derivs)+1)]
        dNvs = [self.bases[1].evaluate(v, d, above) for d in range(np.sum(derivs)+1)]

        d0ud0v = evaluate([dNus[0], dNvs[0]], self.controlpoints, tensor)
        d1ud0v = evaluate([dNus[1], dNvs[0]], self.controlpoints, tensor)
        d0ud1v = evaluate([dNus[0], dNvs[1]], self.controlpoints, tensor)
        d1ud1v = evaluate([dNus[1], dNvs[1]], self.controlpoints, tensor)
        d2ud0v = evaluate([dNus[2], dNvs[0]], self.controlpoints, tensor)
        d0ud2v = evaluate([dNus[0], dNvs[2]], self.controlpoints, tensor)
        W     = d0ud0v[:,:,-1]
        dWdu  = d1ud0v[:,:,-1]
        dWdv  = d0ud1v[:,:,-1]
        d2Wduv= d1ud1v[:,:,-1]
        d2Wdu = d2ud0v[:,:,-1]
        d2Wdv = d0ud2v[:,:,-1]

        for i in range(self.dimension):
            H1   = d1ud0v[:,:,i] * W - d0ud0v[:,:,i] * dWdu
            H2   = d0ud1v[:,:,i] * W - d0ud0v[:,:,i] * dWdv
            dH1du= d2ud0v[:,:,i] * W - d0ud0v[:,:,i] * d2Wdu
            dH1dv= d1ud1v[:,:,i] * W + d1ud0v[:,:,i] * dWdv - d0ud1v[:,:,i] * dWdu - d0ud0v[:,:,i] * d2Wduv
            dH2du= d1ud1v[:,:,i] * W + d0ud1v[:,:,i] * dWdu - d1ud0v[:,:,i] * dWdv - d0ud0v[:,:,i] * d2Wduv
            dH2dv= d0ud2v[:,:,i] * W - d0ud0v[:,:,i] * d2Wdv
            G1   = dH1du*W - 2*H1*dWdu
            G2   = dH2dv*W - 2*H2*dWdv
            if derivs == (1,0):
                result[:,:,i] = H1 / W/W
            elif derivs == (0,1):
                result[:,:,i] = H2 / W/W
            elif derivs == (1,1):
                result[:,:,i] = (dH1dv*W - 2*H1*dWdv) /W/W/W
            elif derivs == (2,0):
                result[:,:,i] = G1 /W/W/W
            elif derivs == (0,2):
                result[:,:,i] = G2 /W/W/W
            if np.sum(derivs) > 2:
                d2ud1v = evaluate([dNus[2], dNvs[1]], self.controlpoints, tensor)
                d1ud2v = evaluate([dNus[1], dNvs[2]], self.controlpoints, tensor)
                d3ud0v = evaluate([dNus[3], dNvs[0]], self.controlpoints, tensor)
                d0ud3v = evaluate([dNus[0], dNvs[3]], self.controlpoints, tensor)
                d3Wdu   = d3ud0v[:,:,-1]
                d3Wdv   = d0ud3v[:,:,-1]
                d3Wduuv = d2ud1v[:,:,-1]
                d3Wduvv = d1ud2v[:,:,-1]
                d2H1du  = d3ud0v[:,:,i]*W + d2ud0v[:,:,i]*dWdu - d1ud0v[:,:,i]*d2Wdu - d0ud0v[:,:,i]*d3Wdu
                d2H1duv = d2ud1v[:,:,i]*W + d2ud0v[:,:,i]*dWdv - d0ud1v[:,:,i]*d2Wdu - d0ud0v[:,:,i]*d3Wduuv
                d2H2dv  = d0ud3v[:,:,i]*W + d0ud2v[:,:,i]*dWdv - d0ud1v[:,:,i]*d2Wdv - d0ud0v[:,:,i]*d3Wdv
                d2H2duv = d1ud2v[:,:,i]*W + d0ud2v[:,:,i]*dWdu - d1ud0v[:,:,i]*d2Wdv - d0ud0v[:,:,i]*d3Wduvv
                dG1du   = d2H1du *W + dH1du*dWdu - 2*dH1du*dWdu - 2*H1*d2Wdu
                dG1dv   = d2H1duv*W + dH1du*dWdv - 2*dH1dv*dWdu - 2*H1*d2Wduv
                dG2du   = d2H2duv*W + dH2dv*dWdu - 2*dH2du*dWdv - 2*H2*d2Wduv
                dG2dv   = d2H2dv *W + dH2dv*dWdv - 2*dH2dv*dWdv - 2*H2*d2Wdv

                if derivs == (3,0):
                    result[:,:,i] = (dG1du*W -3*G1*dWdu) /W/W/W/W
                elif derivs == (0,3):
                    result[:,:,i] = (dG2dv*W -3*G2*dWdv) /W/W/W/W
                elif derivs == (2,1):
                    result[:,:,i] = (dG1dv*W -3*G1*dWdv) /W/W/W/W
                elif derivs == (1,2):
                    result[:,:,i] = (dG2du*W -3*G2*dWdu) /W/W/W/W

        # Squeeze the singleton dimensions if we only have one point
        if squeeze:
            result = result.reshape(self.dimension)

        return result


    def area(self):
        """ Computes the area of the surface in geometric space """
        # fetch integration points
        (x1,w1) = np.polynomial.legendre.leggauss(self.order(0)+1)
        (x2,w2) = np.polynomial.legendre.leggauss(self.order(1)+1)
        # map points to parametric coordinates (and update the weights)
        (knots1,knots2) = self.knots()
        u  = np.array([ (x1+1)/2*(t1-t0)+t0 for t0,t1 in zip(knots1[:-1], knots1[1:]) ])
        w1 = np.array([     w1/2*(t1-t0)    for t0,t1 in zip(knots1[:-1], knots1[1:]) ])
        v  = np.array([ (x2+1)/2*(t1-t0)+t0 for t0,t1 in zip(knots2[:-1], knots2[1:]) ])
        w2 = np.array([     w2/2*(t1-t0)    for t0,t1 in zip(knots2[:-1], knots2[1:]) ])

        # wrap everything to vectors
        u = np.ndarray.flatten(u)
        v = np.ndarray.flatten(v)
        w1 = np.ndarray.flatten(w1)
        w2 = np.ndarray.flatten(w2)

        # compute all quantities of interest (i.e. the jacobian)
        du = self.derivative(u,v, d=(1,0))
        dv = self.derivative(u,v, d=(0,1))
        J  = np.cross(du,dv)

        if self.dimension == 3:
            J = np.sqrt(np.sum(J**2, axis=2))
        else:
            J = np.abs(J)
        return w1.dot(J).dot(w2)

    def edges(self):
        """Return the four edge curves in (parametric) order: umin, umax, vmin, vmax

        :return: Edge curves
        :rtype: (Curve)
        """
        return tuple(self.section(*args) for args in sections(2, 1))

    def const_par_curve(self, knot, direction):
        """  Get a Curve representation of the parametric line of some constant
        knot value.
        :param float knot: The constant knot value to sample the surface
        :param int direction: The parametric direction for the constant value
        :return: curve on this surface
        :rtype: Curve
        """
        direction = check_direction(direction, 2)

        # clone basis since we need to augment this by knot insertion
        b    = self.bases[direction].clone()

        # compute mapping matrix C which is the knotinsertion operator
        mult = min(b.continuity(knot), b.order-1)
        C    = np.identity(self.shape[direction])
        for i in range(mult):
            C = b.insert_knot(knot) @ C

        # at this point we have a C0 basis, find the right interpolating index
        i  = max(bisect_left(b.knots, knot) - 1,0)

        # compute the controlpoints and return Curve
        cp = np.tensordot(C[i,:], self.controlpoints, axes=(0, direction))
        return Curve(self.bases[1-direction], cp, self.rational)

    def rebuild(self, p, n):
        """  Creates an approximation to this surface by resampling it using
        uniform knot vectors of order *p* with *n* control points.

        :param (int) p: Tuple of polynomial discretization order in each direction
        :param (int) n: Tuple of number of control points in each direction
        :return: A new approximate surface
        :rtype: Surface
        """
        p = ensure_listlike(p, dups=2)
        n = ensure_listlike(n, dups=2)

        old_basis = self.bases
        basis = []
        u = []
        N = []
        # establish uniform open knot vectors
        for i in range(2):
            knot = [0] * p[i] + list(range(1, n[i] - p[i] + 1)) + [n[i] - p[i] + 1] * p[i]
            basis.append(BSplineBasis(p[i], knot))

            # make these span the same parametric domain as the old ones
            basis[i].normalize()
            t0 = old_basis[i].start()
            t1 = old_basis[i].end()
            basis[i] *= (t1 - t0)
            basis[i] += t0

            # fetch evaluation points and evaluate basis functions
            u.append(basis[i].greville())
            N.append(basis[i].evaluate(u[i]))

        # find interpolation points as evaluation of existing surface
        x = self.evaluate(u[0], u[1])

        # solve interpolation problem
        cp = np.tensordot(np.linalg.inv(N[1]), x, axes=(1, 1))
        cp = np.tensordot(np.linalg.inv(N[0]), cp, axes=(1, 1))

        # re-order controlpoints so they match up with Surface constructor
        cp = cp.transpose((1, 0, 2))
        cp = cp.reshape(n[0] * n[1], cp.shape[2])

        # return new resampled curve
        return Surface(basis[0], basis[1], cp)

    def __repr__(self):
        result = str(self.bases[0]) + '\n' + str(self.bases[1]) + '\n'
        # print legacy controlpoint enumeration
        n1, n2, n3 = self.controlpoints.shape
        for j in range(n2):
            for i in range(n1):
                result += str(self.controlpoints[i, j, :]) + '\n'
        return result

    get_derivative_surface = SplineObject.get_derivative_spline



__doc__ = 'Implementation of convenience methods with respect to nutils integration.'

import numpy as np

from ..curve import Curve
from ..surface import Surface
from ..volume import Volume


def controlpoints(spline):
    """ Return controlpoints according to nutils ordering """
    n = len(spline)
    dim = spline.dimension
    if isinstance(spline, Curve):
        return np.reshape(spline[:,:]                  , (n, dim), order='F')
    elif isinstance(spline, Surface):
        return np.reshape(spline[:,:,:].swapaxes(0,1)  , (n, dim), order='F')
    elif isinstance(spline, Volume):
        return np.reshape(spline[:,:,:,:].swapaxes(0,2), (n, dim), order='F')
    raise RuntimeError('Non-spline argument detected')

def multiplicities(spline):
    """  Returns the multiplicity of the knots at all knot values as a 2D array for
    all parametric directions, for all knots """
    return [[spline.order(d) - spline.bases[d].continuity(k) - 1 for k in spline.knots(d)] for d in range(spline.pardim)]

def degree(spline):
    """ Returns polynomial degree (splipy order - 1) for all parametric directions """
    return [p-1 for p in spline.order()]

def splipy_to_nutils(spline):
    """ Returns nutils domain and geometry object for spline mapping given by the argument """
    from nutils import mesh, function
    domain, geom = mesh.rectilinear(spline.knots())
    cp    = controlpoints(spline)
    basis = domain.basis('spline', degree=degree(spline), knotmultiplicities=multiplicities(spline))
    geom  = function.matmat(basis, cp)
    #TODO: add correct behaviour for rational and/or periodic geometries
    return domain, geom




# -*- coding: utf-8 -*-

from math import pi

import numpy as np
from scipy.spatial import ConvexHull

from .basis import BSplineBasis
from .curve import Curve
from .surface import Surface
from .splineobject import SplineObject
from .utils import ensure_listlike, check_direction, sections
from . import state

__all__ = ['TrimmedSurface']


class TrimmedSurface(Surface):
    """TrimmedSurface()

    Represents a surface: an object with a two-dimensional parameter space and
    one or more interior closed trimming loops."""

    _intended_pardim = 2

    def __init__(self, basis1=None, basis2=None, controlpoints=None, rational=False, loops=None, **kwargs):
        """  Construct a surface with the given basis and control points.

        The default is to create a linear one-element mapping from and to the
        unit square.

        :param BSplineBasis basis1: The basis of the first parameter direction
        :param BSplineBasis basis2: The basis of the second parameter direction
        :param array-like controlpoints: An *n1* × *n2* × *d* matrix of control points
        :param bool rational: Whether the surface is rational (in which case the
            control points are interpreted as pre-multiplied with the weight,
        :param [[Curve]] Loops: One or more loops. Clockwise loops define holes,
            while counter-clockwise define surface interiors (materials always on the "left")
        :raises RuntimeError: If the loops are not contained to dimension 2 (parametric
            space), or if they are not closed, or if they are not looping properly
        """
        super(Surface, self).__init__([basis1, basis2], controlpoints, rational, **kwargs)
        # make sure to make deep copies of the loops so nothing bad happens
        self.boundaries = [[l.clone() for l in one_loop] for one_loop in loops]

        # error check input curves
        for one_loop in self.boundaries:
            for curve in one_loop:
                if not curve.dimension == 2:
                    raise RuntimeError('Boundary curves need to have dimension 2')
            for i in range(len(one_loop)):
                # print(state.parametric_absolute_tolerance)
                # print(one_loop[i-1][-1,:], ' ', one_loop[i][0,:])
                if not np.allclose(one_loop[i-1][-1,:], one_loop[i][0,:],
                                   rtol=state.parametric_relative_tolerance,
                                   atol=state.parametric_absolute_tolerance):
                    raise RuntimeError('Boundary curves not closed')

        self.__compute_convex_hulls()

    def edges(self):
        """Return the four edge curves in (parametric) order: umin, umax, vmin, vmax

        :return: Edge curves
        :rtype: (Curve)
        """
        return tuple(self.section(*args) for args in sections(2, 1))

    def __compute_convex_hulls(self):
        self.rotation   = []
        self.convexhull = []
        for loop in self.boundaries:
            # don't know if we really need the error test for boundary loops, but
            # I'll just leave it in, because I wrote i and kind of like it

            # get all controlpoints for all the curves, make sure not to double-
            # count the end and start of subsequent curve-pieces
            x = np.vstack([curve[1:,:2] for curve in loop])
            # compute vectors between the control points (velocity approximation?)
            dx = np.diff(np.append(x, [x[0,:]], axis=0), axis=0)
            # compute angle at all points
            theta = np.arctan2(dx[:,1], dx[:,0])
            # compute angle difference at all (control-)points
            dt = np.diff(np.append(theta, theta[0]))
            dt[np.where( dt<-pi) ] += 2*pi
            dt[np.where( dt>+pi) ] -= 2*pi

            total_rotation = np.sum(dt)
            if np.allclose(total_rotation, 2*pi):
                self.rotation.append('counterclockwise')
            elif np.allclose(total_rotation, -2*pi):
                self.rotation.append('clockwise')
            else:
                raise RuntimeError('Boundary loops does not loop exactly once')

            hull = ConvexHull(x)
            self.convexhull.append(x[hull.vertices,:])
            # print(self.convexhull[-1])

    def is_contained(self, u, v):
        """ Returns a boolean mask if the input points are inside (True) or 
        outside (False) of the trimming curves."""

        raise NotImplementedError('This has yet to be implemented')

        # do a quick test based on convex hull
        self.__is_contained_coarse(u,v)
        # for all points that are still undecided, do a fine-grained newton
        # iteration approach
        self.__is_contained_fine(u,v)

        return False

    def __is_contained_fine(self, u, v):
        """ Does a fine test based on parametric curve representation to see if
        points are inside or outside trimming domain. Trimming curves are high-
        polynomial representations, so figuring this out means newton iteration
        to locate nearest point on curve and decide if this is inside or outside
        domain."""
        return False

    def __is_contained_coarse(self, u, v):
        """ Does a course test based on control-grid to see if points are inside or
        outside domain. Inside control-grid means inside a trimming loop and outputs
        False. Outside the *convex hull* of a control-grid means"""
        return False




# -*- coding: utf-8 -*-

"""Handy utilities for creating surfaces."""

from math import pi, sqrt, atan2
import inspect
import os
from os.path import dirname, realpath, join

import numpy as np

from .basis import BSplineBasis
from .curve import Curve
from .surface import Surface
from .utils import flip_and_move_plane_geometry, rotate_local_x_axis
from .utils.nutils import controlpoints, multiplicities, degree
from . import curve_factory, state

__all__ = ['square', 'disc', 'sphere', 'extrude', 'revolve', 'cylinder', 'torus', 'edge_curves',
           'thicken', 'sweep', 'loft', 'interpolate', 'least_square_fit', 'teapot']


def square(size=1, lower_left=(0,0)):
    """  Create a square with parametric origin at *(0,0)*.

    :param float size: Size(s), either a single scalar or a tuple of scalars per axis
    :param array-like lower_left: local origin, the lower left corner of the square
    :return: A linear parametrized square
    :rtype: Surface
    """
    result = Surface()  # unit square
    result.scale(size)
    result += lower_left
    return result


def disc(r=1, center=(0,0,0), normal=(0,0,1), type='radial', xaxis=(1,0,0)):
    """  Create a circular disc. The *type* parameter distinguishes between
    different parametrizations.

    :param float r: Radius
    :param array-like center: local origin
    :param array-like normal: local normal
    :param string type: The type of parametrization ('radial' or 'square')
    :param array-like xaxis: direction of sem, i.e. parametric start point v=0
    :return: The disc
    :rtype: Surface
    """
    if type == 'radial':
        c1 = curve_factory.circle(r, center=center, normal=normal, xaxis=xaxis)
        c2 = flip_and_move_plane_geometry(c1*0, center, normal)
        result = edge_curves(c2, c1)
        result.swap()
        result.reparam((0,r), (0,2*pi))
        return result
    elif type == 'square':
        w = 1 / sqrt(2)
        cp = [[-r * w, -r * w, 1],
              [0, -r, w],
              [r * w, -r * w, 1],
              [-r, 0, w],
              [0, 0, 1],
              [r, 0, w],
              [-r * w, r * w, 1],
              [0, r, w],
              [r * w, r * w, 1]]
        basis1 = BSplineBasis(3)
        basis2 = BSplineBasis(3)
        result = Surface(basis1, basis2, cp, True)
        return flip_and_move_plane_geometry(result, center, normal)
    else:
        raise ValueError('invalid type argument')

def sphere(r=1, center=(0,0,0), zaxis=(0,0,1), xaxis=(1,0,0)):
    """  Create a spherical shell.

    :param float r: Radius
    :param array-like center: Local origin of the sphere
    :param array-like zaxis: direction of the north/south pole of the parametrization
    :param array-like xaxis: direction of the longitudal sem
    :return: The spherical shell
    :rtype: Surface
    """
    circle = curve_factory.circle_segment(pi, r)
    circle.rotate(-pi / 2)
    circle.rotate(pi / 2, (1, 0, 0))  # flip up into xz-plane
    result = revolve(circle)

    result.rotate(rotate_local_x_axis(xaxis, zaxis))
    return flip_and_move_plane_geometry(result, center, zaxis)


def extrude(curve, amount):
    """  Extrude a curve by sweeping it to a given height.

    :param Curve curve: Curve to extrude
    :param array-like amount: 3-component vector of sweeping amount and
                               direction
    :return: The extruded curve
    :rtype: Surface
    """
    curve = curve.clone()  # clone input curve, throw away input reference
    curve.set_dimension(3)  # add z-components (if not already present)
    n = len(curve)  # number of control points of the curve
    cp = np.zeros((2 * n, curve.dimension + curve.rational))
    cp[:n, :] = curve.controlpoints  # the first control points form the bottom
    curve += amount
    cp[n:, :] = curve.controlpoints  # the last control points form the top
    return Surface(curve.bases[0], BSplineBasis(2), cp, curve.rational)


def revolve(curve, theta=2 * pi, axis=(0,0,1)):
    """  Revolve a surface by sweeping a curve in a rotational fashion around
    the *z* axis.

    :param Curve curve: Curve to revolve
    :param float theta: Angle to revolve, in radians
    :param array-like axis: Axis of rotation
    :return: The revolved surface
    :rtype: Surface
    """
    curve = curve.clone()  # clone input curve, throw away input reference
    curve.set_dimension(3)  # add z-components (if not already present)
    curve.force_rational()  # add weight (if not already present)

    # align axis with the z-axis
    normal_theta = atan2(axis[1], axis[0])
    normal_phi   = atan2(sqrt(axis[0]**2 + axis[1]**2), axis[2])
    curve.rotate(-normal_theta, [0,0,1])
    curve.rotate(-normal_phi,   [0,1,0])

    circle_seg = curve_factory.circle_segment(theta)

    n = len(curve)      # number of control points of the curve
    m = len(circle_seg) # number of control points of the sweep
    cp = np.zeros((m * n, 4))

    # loop around the circle and set control points by the traditional 9-point
    # circle curve with weights 1/sqrt(2), only here C0-periodic, so 8 points
    dt = 0
    t  = 0
    for i in range(m):
        x,y,w = circle_seg[i]
        dt  = atan2(y,x) - t
        t  += dt
        curve.rotate(dt)
        cp[i * n:(i + 1) * n, :]  = curve[:]
        cp[i * n:(i + 1) * n, 2] *= w
        cp[i * n:(i + 1) * n, 3] *= w
    result = Surface(curve.bases[0], circle_seg.bases[0], cp, True)
    # rotate it back again
    result.rotate(normal_phi,   [0,1,0])
    result.rotate(normal_theta, [0,0,1])
    return result


def cylinder(r=1, h=1, center=(0,0,0), axis=(0,0,1), xaxis=(1,0,0)):
    """  Create a cylinder shell with no top or bottom

    :param float r: Radius
    :param float h: Height
    :param array-like center: The center of the bottom circle
    :param array-like axis: Cylinder axis
    :param array-like xaxis: direction of sem, i.e. parametric start point u=0
    :return: The cylinder shell
    :rtype: Surface
    """
    return extrude(curve_factory.circle(r, center, axis, xaxis=xaxis), h*np.array(axis))


def torus(minor_r=1, major_r=3, center=(0,0,0), normal=(0,0,1), xaxis=(1,0,0)):
    """  Create a torus (doughnut) by revolving a circle of size *minor_r*
    around the *z* axis with radius *major_r*.

    :param float minor_r: The thickness of the torus (radius in the *xz* plane)
    :param float major_r: The size of the torus (radius in the *xy* plane)
    :param array-like center: Local origin of the torus
    :param array-like normal: Local origin of the torus
    :param array-like center: Local origin of the torus
    :return: A periodic torus
    :rtype: Surface
    """
    circle = curve_factory.circle(minor_r)
    circle.rotate(pi / 2, (1, 0, 0))  # flip up into xz-plane
    circle.translate((major_r, 0, 0))  # move into position to spin around z-axis
    result = revolve(circle)

    result.rotate(rotate_local_x_axis(xaxis, normal))
    return flip_and_move_plane_geometry(result, center, normal)


def edge_curves(*curves, **kwargs):
    """  Create the surface defined by the region between the input curves.

    In case of four input curves, these must be given in an ordered directional
    closed loop around the resulting surface.

    :param [Curve] curves: Two or four edge curves
    :param string type: The method used for interior computation ('coons', 'poisson', 'elasticity' or 'finitestrain')
    :return: The enclosed surface
    :rtype: Surface
    :raises ValueError: If the length of *curves* is not two or four
    """
    type = kwargs.get('type', 'coons')
    if len(curves) == 1: # probably gives input as a list-like single variable
        curves = curves[0]
    if len(curves) == 2:
        crv1 = curves[0].clone()
        crv2 = curves[1].clone()
        Curve.make_splines_identical(crv1, crv2)
        (n, d) = crv1.controlpoints.shape  # d = dimension + rational

        controlpoints = np.zeros((2 * n, d))
        controlpoints[:n, :] = crv1.controlpoints
        controlpoints[n:, :] = crv2.controlpoints
        linear = BSplineBasis(2)

        return Surface(crv1.bases[0], linear, controlpoints, crv1.rational)
    elif len(curves) == 4:
        # reorganize input curves so they form a directed loop around surface
        rtol = state.controlpoint_relative_tolerance
        atol = state.controlpoint_absolute_tolerance
        mycurves = [c.clone() for c in curves] # wrap into list and clone all since we're changing them
        dim = np.max([c.dimension for c in mycurves])
        rat = np.any([c.rational  for c in mycurves])
        for i in range(4):
            for j in range(i+1,4):
                Curve.make_splines_compatible(mycurves[i], mycurves[j])

        if not (np.allclose(mycurves[0][-1], mycurves[1][0], rtol=rtol, atol=atol) and
                np.allclose(mycurves[1][-1], mycurves[2][0], rtol=rtol, atol=atol) and
                np.allclose(mycurves[2][-1], mycurves[3][0], rtol=rtol, atol=atol) and
                np.allclose(mycurves[3][-1], mycurves[0][0], rtol=rtol, atol=atol)):
            reorder = [mycurves[0]]
            del mycurves[0]
            for j in range(3):
                found_match = False
                for i in range(len(mycurves)):
                    if(np.allclose(reorder[j][-1], mycurves[i][0], rtol=rtol, atol=atol)):
                        reorder.append(mycurves[i])
                        del mycurves[i]
                        found_match = True
                        break
                    elif(np.allclose(reorder[j][-1], mycurves[i][-1], rtol=rtol, atol=atol)):
                        reorder.append(mycurves[i].reverse())
                        del mycurves[i]
                        found_match = True
                        break
                if not found_match:
                    raise RuntimeError('Curves do not form a closed loop (end-points do not match)')
            mycurves = reorder
        if type == 'coons':
            return coons_patch(*mycurves)
        elif type == 'poisson':
            return poisson_patch(*mycurves)
        elif type == 'elasticity':
            return elasticity_patch(*mycurves)
        elif type == 'finitestrain':
            return finitestrain_patch(*mycurves)
        else:
            raise ValueError('Unknown type parameter')
    else:
        raise ValueError('Requires two or four input curves')


def coons_patch(bottom, right, top, left):
    """  Create the surface defined by the region between the 4 input curves.

    The input curves need to be parametrized to form a directed loop around the resulting Surface.
    For more information on Coons patch see: https://en.wikipedia.org/wiki/Coons_patch.

    :param [Curve] bottom: curve corresponding to the result parametric value v=0
    :param [Curve] right: curve corresponding to the result parametric value u=1
    :param [Curve] top: curve corresponding to the result parametric value v=1 (reversed: going right-left)
    :param [Curve] left: curve corresponding to the result parametric value u=0 (reversed: going top-bottom)
    :return: The enclosed surface
    :rtype: Surface
    """
    # coons patch (https://en.wikipedia.org/wiki/Coons_patch)
    top.reverse()
    left.reverse()

    # create linear interpolation between opposing sides
    s1 = edge_curves(bottom, top)
    s2 = edge_curves(left, right)
    s2.swap()
    # create (linear,linear) corner parametrization
    linear = BSplineBasis(2)
    rat = s1.rational  # using control-points from top/bottom, so need to know if these are rational
    if rat:
        bottom = bottom.clone().force_rational() # don't mess with the input curve, make clone
        top.force_rational()                     # this is already a clone
    s3 = Surface(linear, linear, [bottom[0], bottom[-1], top[0], top[-1]], rat)

    # in order to add spline surfaces, they need identical parametrization
    Surface.make_splines_identical(s1, s2)
    Surface.make_splines_identical(s1, s3)
    Surface.make_splines_identical(s2, s3)

    result = s1
    result.controlpoints += s2.controlpoints
    result.controlpoints -= s3.controlpoints
    return result


def poisson_patch(bottom, right, top, left):
    from nutils import version
    if int(version[0]) != 4:
        raise ImportError('Mismatching nutils version detected, only version 4 supported. Upgrade by \"pip install --upgrade nutils\"')

    from nutils import mesh, function as fn
    from nutils import _, log

    # error test input
    if left.rational or right.rational or top.rational or bottom.rational:
        raise RuntimeError('poisson_patch not supported for rational splines')

    # these are given as a oriented loop, so make all run in positive parametric direction
    top.reverse()
    left.reverse()

    # in order to add spline surfaces, they need identical parametrization
    Curve.make_splines_identical(top, bottom)
    Curve.make_splines_identical(left, right)

    # create computational (nutils) mesh
    p1 = bottom.order(0)
    p2 = left.order(0)
    n1 = len(bottom)
    n2 = len(left)
    dim= left.dimension
    k1 = bottom.knots(0)
    k2 = left.knots(0)
    m1 = [bottom.order(0) - bottom.continuity(k) - 1 for k in k1]
    m2 = [left.order(0)   - left.continuity(k)   - 1 for k in k2]
    domain, geom = mesh.rectilinear([k1, k2])
    basis = domain.basis('spline', [p1-1, p2-1], knotmultiplicities=[m1,m2])

    # assemble system matrix
    grad      = basis.grad(geom)
    outer     = fn.outer(grad,grad)
    integrand = outer.sum(-1)
    matrix = domain.integrate(integrand * fn.J(geom), ischeme='gauss'+str(max(p1,p2)+1))

    # initialize variables
    controlpoints = np.zeros((n1,n2,dim))
    rhs           = np.zeros((n1*n2))
    constraints = np.array([[np.nan]*n2]*n1)

    # treat all dimensions independently
    for d in range(dim):
        # add boundary conditions
        constraints[ 0, :] = left[  :,d]
        constraints[-1, :] = right[ :,d]
        constraints[ :, 0] = bottom[:,d]
        constraints[ :,-1] = top[   :,d]

        # solve system
        lhs = matrix.solve(rhs, constrain=np.ndarray.flatten(constraints))

        # wrap results into splipy datastructures
        controlpoints[:,:,d] = np.reshape(lhs, (n1,n2), order='C')

    return Surface(bottom.bases[0], left.bases[0], controlpoints, bottom.rational, raw=True)


def elasticity_patch(bottom, right, top, left):
    from nutils import version
    if int(version[0]) != 4:
        raise ImportError('Mismatching nutils version detected, only version 4 supported. Upgrade by \"pip install --upgrade nutils\"')

    from nutils import mesh, function, solver
    from nutils import _, log

    # error test input
    if not (left.dimension == right.dimension == top.dimension == bottom.dimension == 2):
        raise RuntimeError('elasticity_patch only supported for planar (2D) geometries')
    if left.rational or right.rational or top.rational or bottom.rational:
        raise RuntimeError('elasticity_patch not supported for rational splines')

    # these are given as a oriented loop, so make all run in positive parametric direction
    top.reverse()
    left.reverse()

    # in order to add spline surfaces, they need identical parametrization
    Curve.make_splines_identical(top, bottom)
    Curve.make_splines_identical(left, right)

    # create computational (nutils) mesh
    p1 = bottom.order(0)
    p2 = left.order(0)
    n1 = len(bottom)
    n2 = len(left)
    dim= left.dimension
    k1 = bottom.knots(0)
    k2 = left.knots(0)
    m1 = [bottom.order(0) - bottom.continuity(k) - 1 for k in k1]
    m2 = [left.order(0)   - left.continuity(k)   - 1 for k in k2]
    domain, geom = mesh.rectilinear([k1, k2])

    # assemble system matrix
    ns = function.Namespace()
    ns.x = geom
    ns.basis = domain.basis('spline', degree=[p1-1,p2-1], knotmultiplicities=[m1,m2]).vector(2)
    ns.eye = np.array([[1,0],[0,1]])
    ns.lmbda = 1
    ns.mu = .3
    # ns.u_i = 'basis_ni ?lhs_n'
    # ns.strain_ij = '(u_i,j + u_j,i) / 2'
    # ns.stress_ij = 'lmbda strain_kk eye_ij + 2 mu strain_ij'
    ns.strain_nij = '(basis_ni,j + basis_nj,i) / 2'
    ns.stress_nij = 'lmbda strain_nkk eye_ij + 2 mu strain_nij'

    # construct matrix and right hand-side
    matrix = domain.integrate(ns.eval_nm('strain_nij stress_mij d:x'), ischeme='gauss'+str(max(p1,p2)+1))
    rhs    = np.zeros((n1*n2*dim))

    # add boundary conditions
    constraints = np.array([[[np.nan]*n2]*n1]*dim)
    for d in range(dim):
        constraints[d, 0, :] = left[  :,d]
        constraints[d,-1, :] = right[ :,d]
        constraints[d, :, 0] = bottom[:,d]
        constraints[d, :,-1] = top[   :,d]

    # solve system
    lhs = matrix.solve(rhs, constrain=np.ndarray.flatten(constraints, order='C'))

    # rewrap results into splipy datastructures
    controlpoints = np.reshape(lhs, (dim,n1,n2), order='C')
    controlpoints = controlpoints.swapaxes(0,2)
    controlpoints = controlpoints.swapaxes(0,1)

    return Surface(bottom.bases[0], left.bases[0], controlpoints, bottom.rational, raw=True)


def finitestrain_patch(bottom, right, top, left):
    from nutils import version
    if int(version[0]) != 4:
        raise ImportError('Mismatching nutils version detected, only version 4 supported. Upgrade by \"pip install --upgrade nutils\"')

    from nutils import mesh, function
    from nutils import _, log, solver

    # error test input
    if not (left.dimension == right.dimension == top.dimension == bottom.dimension == 2):
        raise RuntimeError('finitestrain_patch only supported for planar (2D) geometries')
    if left.rational or right.rational or top.rational or bottom.rational:
        raise RuntimeError('finitestrain_patch not supported for rational splines')

    # these are given as a oriented loop, so make all run in positive parametric direction
    top.reverse()
    left.reverse()

    # in order to add spline surfaces, they need identical parametrization
    Curve.make_splines_identical(top, bottom)
    Curve.make_splines_identical(left, right)

    # create an initial mesh (correct corners) which we will morph into the right one
    p1 = bottom.order(0)
    p2 = left.order(0)
    p  = max(p1,p2)
    linear = BSplineBasis(2)
    srf = Surface(linear, linear, [bottom[0], bottom[-1], top[0], top[-1]])
    srf.raise_order(p1-2, p2-2)
    for k in bottom.knots(0, True)[p1:-p1]:
        srf.insert_knot(k, 0)
    for k in left.knots(0, True)[p2:-p2]:
        srf.insert_knot(k, 1)

    # create computational mesh
    n1 = len(bottom)
    n2 = len(left)
    dim= left.dimension
    domain, geom = mesh.rectilinear(srf.knots())
    ns = function.Namespace()
    ns.basis = domain.basis('spline', degree(srf), knotmultiplicities=multiplicities(srf)).vector( 2 )
    ns.phi   = domain.basis('spline', degree(srf), knotmultiplicities=multiplicities(srf))
    ns.eye       = np.array([[1,0],[0,1]])
    ns.cp        = controlpoints(srf)
    ns.x_i       = 'cp_ni phi_n'
    ns.lmbda     = 1
    ns.mu        = 1

    # add total boundary conditions
    # for hard problems these will be taken in steps and multiplied by dt every
    # time (quasi-static iterations)
    constraints = np.array([[[np.nan]*n2]*n1]*dim)
    for d in range(dim):
        constraints[d, 0, :] = (left[  :,d] - srf[ 0, :,d])
        constraints[d,-1, :] = (right[ :,d] - srf[-1, :,d])
        constraints[d, :, 0] = (bottom[:,d] - srf[ :, 0,d])
        constraints[d, :,-1] = (top[   :,d] - srf[ :,-1,d])
    # TODO: Take a close look at the logic below

    # in order to iterate, we let t0=0 be current configuration and t1=1 our target configuration
    # if solver divergeces (too large deformation), we will try with dt=0.5. If this still
    # fails we will resort to dt=0.25 until a suitable small iterations size have been found

    # dt = 1
    # t0 = 0
    # t1 = 1
    # while t0 < 1:
        # dt = t1-t0
    n  = 10
    dt = 1/n
    for i in range(n):
        # print(' ==== Quasi-static '+str(t0*100)+'-'+str(t1*100)+' % ====')
        print(' ==== Quasi-static '+str(i/(n-1)*100)+' % ====')

        # define the non-linear finite strain problem formulation
        ns.cp        = np.reshape(srf[:,:,:].swapaxes(0,1), (n1*n2, dim), order='F')
        ns.x_i       = 'cp_ni phi_n'   # geometric mapping (reference geometry)
        ns.u_i       = 'basis_ki ?w_k' # displacement (unknown coefficients w_k)
        ns.X_i       = 'x_i + u_i'     # displaced geometry
        ns.strain_ij = '.5 (u_i,j + u_j,i + u_k,i u_k,j)'
        ns.stress_ij = 'lmbda strain_kk eye_ij + 2 mu strain_ij'

        # try:
        residual = domain.integral(ns.eval_n('stress_ij basis_ni,j d:X'), degree=2*p)
        cons = np.ndarray.flatten(constraints*dt, order='C')
        lhs = solver.newton('w', residual, constrain=cons).solve(
                            tol=state.controlpoint_absolute_tolerance, maxiter=8)

        # store the results on a splipy object and continue
        geom   = lhs.reshape((n2,n1,dim), order='F')
        srf[:,:,:] += geom.swapaxes(0,1)

        # t0 += dt
        # t1 = 1
        # except solver.SolverError: # newton method fail to converge, try a smaller step length 'dt'
            # t1 = (t1+t0)/2
    return srf

def thicken(curve, amount):
    """  Generate a surface by adding thickness to a curve.

    - For 2D curves this will generate a 2D planar surface with the curve
      through the center.

    - For 3D curves this will generate a surface "tube" which is open at both
      ends (that is, for a straight line it will generate a cylinder shell).

    The resulting surface is an approximation generated by interpolating at the
    Greville points. It will use the same discretization as the input curve.
    It does not check for self-intersecting geometries.

    :param Curve curve: The generating curve
    :param amount: The amount of thickness, either constant or variable (if
        variable, the function must accept parameters named *x*, *y*, *z* and/or *t*)
    :return: Surrounding surface
    :rtype: Surface
    """
    # NOTES: There are several pitfalls with this function
    #  * self intersection:
    #     could be handled more gracefully, but is here ignored
    #  * choice of discretization:
    #     the offset curve is computed from the velocity (tangent) which is of
    #     one continuity less than the original curve. In particular C1
    #     quadratic curves will get very noticeable C0-kinks in them. Currently
    #     this is completely ignored and we keep the high continuity of the
    #     original curve.
    #  * width given by function input
    #     could produce wild behaviour. Original discretization might not
    #     produce a satisfactory result

    curve = curve.clone()  # clone input curve, throw away input reference
    t = curve.bases[0].greville()
    if curve.dimension == 2:
        # linear parametrization across domain
        n = len(curve)
        left_points = np.zeros((n, 2))
        right_points = np.zeros((n, 2))
        linear = BSplineBasis(2)

        x = curve.evaluate(t)                 # curve at interpolation points
        v = curve.derivative(t)               # velocity at interpolation points
        l = np.sqrt(v[:, 0]**2 + v[:, 1]**2)  # normalizing factor for velocity
        for i in range(n):
            if l[i] < 1e-13: # in case of zero velocity, use neighbour instead
                if i>0:
                    v[i,:] = v[i-1,:]
                else:
                    v[i,:] = v[i+1,:]
            else:
                v[i,:] /= l[i]

        if inspect.isfunction(amount):
            arg_names = inspect.signature(amount).parameters
            argc = len(arg_names)
            argv = [0] * argc
            for i in range(n):
                # build up the list of arguments (in case not all of (x,y,t) are specified)
                for j,name in enumerate(arg_names):
                    if name == 'x':
                        argv[j] = x[i, 0]
                    elif name == 'y':
                        argv[j] = x[i, 1]
                    elif name == 'z':
                        argv[j] = 0.0
                    elif name == 't':
                        argv[j] = t[i]
                # figure out the distane at this particular point
                dist = amount(*argv)

                # store interpolation points
                right_points[i, 0] = x[i, 0] - v[i, 1] * dist  # x at bottom
                right_points[i, 1] = x[i, 1] + v[i, 0] * dist  # y at bottom
                left_points[ i, 0] = x[i, 0] + v[i, 1] * dist  # x at top
                left_points[ i, 1] = x[i, 1] - v[i, 0] * dist  # y at top
        else:
            right_points[:, 0] = x[:, 0] - v[:, 1] * amount  # x at bottom
            right_points[:, 1] = x[:, 1] + v[:, 0] * amount  # y at bottom
            left_points[ :, 0] = x[:, 0] + v[:, 1] * amount  # x at top
            left_points[ :, 1] = x[:, 1] - v[:, 0] * amount  # y at top
        # perform interpolation on each side
        right = curve_factory.interpolate(right_points, curve.bases[0])
        left  = curve_factory.interpolate(left_points,  curve.bases[0])
        return edge_curves(right, left)

    else:  # dimension=3, we will create a surrounding tube
        return sweep(curve, curve_factory.circle(r=amount))

def sweep(path, shape):
    """  Generate a surface by sweeping a shape along a path

    The resulting surface is an approximation generated by interpolating at the
    Greville points. It is generated by sweeping a shape curve along a path.

    The *shape* object has to be contained in the 'xy' plane (preferably centered
    around the origin) as its x-coordinate is extruded in the normal direction,
    and its y-coordinate in the binormal direction of the *path* curve.

    :param Curve path:  The path to drag *shape* along
    :param Curve shape: The shape to be dragged out to a surface
    :return: Surrounding surface
    :rtype: Surface
    """
    b1 = path.bases[0]
    b2 = shape.bases[0]
    n1 = b1.num_functions()
    n2 = b2.num_functions()
    # this requires binormals and normals, which only work in 3D, so assume this here
    X  = np.zeros((n1,n2, 3))
    for i in range(n1):
        u = b1.greville(i)
        x = path(u)
        B = path.binormal(u)
        N = path.normal(u)
        for j in range(n2):
            v = b2.greville(j)
            y = shape(v)
            X[i,j,:] = x + N*y[0] + B*y[1]

    return interpolate(X, [b1,b2])


def loft(*curves):
    """  Generate a surface by lofting a series of curves

    The resulting surface is interpolated at all input curves and a smooth transition
    between these curves is computed as a cubic spline interpolation in the lofting
    direction. In the case that insufficient curves are provided as input (less than 4
    curves), then a quadratic or linear interpolation is performed.

    Note that the order of input curves matter as they will be interpolated in this
    particular order. Also note that the curves need to be parametrized in the same
    direction, otherwise you will encounter self-intersecting result surface as it is
    wrapping in on itself.

    :param Curve curves:  A sequence of curves to be lofted
    :return: Lofted surface
    :rtype: Surface

    Examples:

    .. code:: python

        from splipy import curve_factory, surface_factory
        crv1 = curve_factory.circle(r=1)
        crv2 = 1.5 * crv1 + [0,0,1]
        crv3 = 1.0 * crv1 + [0,0,2]
        crv4 = 1.3 * crv1 + [0,0,4]
        srf = surface_factory.loft(crv1, crv2, crv3, crv4)

        # alternatively you can provide all input curves as a list
        all_my_curves = [crv1, crv2, crv3, crv4]
        srf = surface_factory.loft(all_my_curves)

    """
    if len(curves) == 1:
        curves = curves[0]

    # clone input, so we don't change those references
    # make sure everything has the same dimension since we need to compute length
    curves = [c.clone().set_dimension(3) for c in curves]
    if len(curves)==2:
        return edge_curves(curves)
    elif len(curves)==3:
        # can't do cubic spline interpolation, so we'll do quadratic
        basis2 = BSplineBasis(3)
        dist  = basis2.greville()
    else:
        x = [c.center() for c in curves]

        # create knot vector from the euclidian length between the curves
        dist = [0]
        for (x1,x0) in zip(x[1:],x[:-1]):
            dist.append(dist[-1] + np.linalg.norm(x1-x0))

        # using "free" boundary condition by setting N'''(u) continuous at second to last and second knot
        knot = [dist[0]]*4 + dist[2:-2] + [dist[-1]]*4
        basis2 = BSplineBasis(4, knot)

    n = len(curves)
    for i in range(n):
        for j in range(i+1,n):
            Curve.make_splines_identical(curves[i], curves[j])

    basis1 = curves[0].bases[0]
    m      = basis1.num_functions()
    u      = basis1.greville() # parametric interpolation points
    v      = dist              # parametric interpolation points

    # compute matrices
    Nu     = basis1(u)
    Nv     = basis2(v)
    Nu_inv = np.linalg.inv(Nu)
    Nv_inv = np.linalg.inv(Nv)

    # compute interpolation points in physical space
    x      = np.zeros((m,n, curves[0][0].size))
    for i in range(n):
        x[:,i,:] = Nu @ curves[i].controlpoints

    # solve interpolation problem
    cp = np.tensordot(Nv_inv, x,  axes=(1,1))
    cp = np.tensordot(Nu_inv, cp, axes=(1,1))

    # re-order controlpoints so they match up with Surface constructor
    cp = cp.transpose((1, 0, 2))
    cp = cp.reshape(n*m, cp.shape[2])

    return Surface(basis1, basis2, cp, curves[0].rational)


def interpolate(x, bases, u=None):
    """  Interpolate a surface on a set of regular gridded interpolation points `x`.

    The points can be either a matrix (in which case the first index is
    interpreted as a flat row-first index of the interpolation grid) or a 3D
    tensor. In both cases the last index is the physical coordinates.

    :param numpy.ndarray x: Grid of interpolation points
    :param [BSplineBasis] bases: The basis to interpolate on
    :param [array-like] u: Parametric interpolation points, defaults to
        Greville points of the basis
    :return: Interpolated surface
    :rtype: Surface
    """
    surf_shape = [b.num_functions() for b in bases]
    dim = x.shape[-1]
    if len(x.shape) == 2:
        x = x.reshape(surf_shape + [dim])
    if u is None:
        u = [b.greville() for b in bases]
    N_all = [b(t) for b,t in zip(bases, u)]
    N_all.reverse()
    cp = x
    for N in N_all:
        cp = np.tensordot(np.linalg.inv(N), cp, axes=(1,1))

    return Surface(bases[0], bases[1], cp.transpose(1,0,2).reshape((np.prod(surf_shape),dim)))


def least_square_fit(x, bases, u):
    """  Perform a least-square fit of a point cloud `x` onto a spline basis.

    The points can be either a matrix (in which case the first index is
    interpreted as a flat row-first index of the interpolation grid) or a 3D
    tensor. In both cases the last index is the physical coordinates.

    There must be at least as many points as basis functions.

    :param numpy.ndarray x: Grid of evaluation points
    :param [BSplineBasis] bases: Basis on which to interpolate
    :param [array-like] u: Parametric values at evaluation points
    :return: Approximated surface
    :rtype: Surface
    """
    surf_shape = [b.num_functions() for b in bases]
    dim = x.shape[-1]
    if len(x.shape) == 2:
        x = x.reshape(surf_shape + [dim])
    N_all = [b(t) for b,t in zip(bases, u)]
    N_all.reverse()
    cp = x
    for N in N_all:
        cp = np.tensordot(N.T, cp, axes=(1,1))
    for N in N_all:
        cp = np.tensordot(np.linalg.inv(N.T @ N), cp, axes=(1,1))

    return Surface(bases[0], bases[1], cp.transpose(1,0,2).reshape((np.prod(surf_shape),dim)))


def teapot():
    """  Generate the Utah teapot as 32 cubic bezier patches. This teapot has a
    rim, but no bottom. It is also self-intersecting making it unsuitable for
    perfect-match multipatch modeling.

    The data is picked from http://www.holmes3d.net/graphics/teapot/

    :return: The utah teapot
    :rtype: List of Surface
    """
    path = join(dirname(realpath(__file__)), 'templates', 'teapot.bpt')
    with open(path) as f:
        results = []
        numb_patches = int(f.readline())
        for i in range(numb_patches):
            p = np.fromstring(f.readline(), dtype=np.uint8, count=2, sep=' ')
            basis1 = BSplineBasis(p[0]+1)
            basis2 = BSplineBasis(p[1]+1)

            ncp = basis1.num_functions() * basis2.num_functions()
            cp  = [np.fromstring(f.readline(), dtype=np.float, count=3, sep=' ') for j in range(ncp)]
            results.append(Surface(basis1, basis2, cp))

    return results



from .g2  import G2
from .svg import SVG
from .spl import SPL
from .stl import STL
from .ofoam import OpenFOAM


# GRDECL depends on optional cv2
try:
    from .grdecl import GRDECL
    has_grdecl = True
except ImportError:
    has_grdecl = False


# ThreeDM depends on optional rhino3dm
try:
    from .threedm import ThreeDM
    has_rhino = True
except ImportError:
    has_rhino = False


__all__ = ['G2', 'SVG', 'SPL', 'STL', 'OpenFOAM']

if has_grdecl:
    __all__.append('GRDECL')

if has_rhino:
    __all__.append('ThreeDM')
